#ifndef __BUF_H__
#define __BUF_H__

#include "pbmtypes.h"
#ifdef NT
#pragma pack(1)
#endif
#define BUF_BLKSIZE         512
#ifdef NT
#ifdef WIN95_ONLY   /* modified by Jeff 2001/11/16 */
#define BUF_MAX_TRK_ITEM    1
#else
#define BUF_MAX_TRK_ITEM    512
#endif
#else
#define BUF_MAX_TRK_ITEM    512     
#endif
/* These flags are used both in buf_req_t && trk_tab_item_t */
#define BUF_FLAG_BYPASS             0x0001
#define BUF_FLAG_READ               0x0002
#define BUF_FLAG_WRITE              0x0004
#define BUF_FLAG_FREE               0x0008

/*#define BUF_FLAG_INVALID_BUF        0x0010  content in buf is invalid */
#define BUF_FLAG_STRIPE_INVALID        0x0010

#define BUF_FLAG_LOCKED             0x0020
#define BUF_FLAG_INTERNAL           0x0040
#define BUF_FLAG_DIRTY              0x0080
#define BUF_FLAG_QUEUED_REQ         0x0100
#define BUF_FLAG_PREREAD            0x0200
#define BUF_FLAG_PREREAD_ONLY       0x0400

/*#define BUF_FLAG_DISCARD            0x0800*/
#define BUF_FLAG_BUF_INVALID		0x0800

#define BUF_FLAG_DO_INVALID_SLOT    0x1000 /* after flush, invalid the slot */
#define BUF_FLAG_FLUSH              0x2000
#define BUF_FLAG_STRIPE_FLUSH       0x4000 /* if set, stripe boundary flush */
#define BUF_WITH_ORIGIN_REQ         0x8000

#define BUF_STATUS_SUSPEND          0x0001
#define BUF_STATUS_NOREQ            0x0002
#define BUF_STATUS_LOCKED           0x0004
#define BUF_STATUS_REQ_DISCARD      0x0008
#define BUF_STATUS_NO_FREE_SLOT     0x0010
#define BUF_STATUS_NO_FREE_BUF      0x0020

#define MAX_SG_COUNT				0x20

#define U32NULL 0xFFFFFFFF

#define FreeNow 0x1
#define UsedNow 0x2
#define backed  0x4

typedef struct _buf_req_t {
    u32                 dev_id;
    baddr_t             blkno;
    bsize_t             size;
	bsize_t				actualSize;
	SG					*sg;
    int                 sg_cnt;
    SG					buf_sg[MAX_SG_COUNT];
    int                 buf_sg_cnt;
    buf_flag_t          flag;
    int                 slot_no;
    void                *ptr;
#ifndef I2O
    U8					p_sns[14];
#endif
    struct _buf_req_t   *pre_read_req;
    struct _buf_req_t   *next;  

#ifdef NT_DEBUG
	unsigned char debugging;
#endif
	
} buf_req_t;

typedef struct _buf_head_t {
    struct _buf_head_t  *prev;
    struct _buf_head_t  *next;
    caddr_t             buf;   
} buf_head_t;

typedef struct {
    u32         dev_id;
    u32         lru;
    u32         lfu;
    buf_flag_t  flag;
    baddr_t     stripe_start;       /* start of the stripe */
    baddr_t     stripe_end;         /* end of the stripe */
    baddr_t     buf_start;          /* start of data in buf */
    baddr_t     buf_end;            /* end of data in buf */
    baddr_t     req_last;            /* last block r/w request accessed */
 
    u32         buf_cnt;            /* how many buffers? */    
    buf_head_t  *buf_head;
    buf_req_t   *request;
    
	u32 lockCount;
	u32 QueCount;
	u32 sequence;
	u32 preReading;

	/*== for performance*/
	baddr_t		virtualBufStart;
	baddr_t		virtualBufEnd;

/*--	int fullStripeFlushing;*/

/* hammer 1120*/
	u32 old_parity;
  	u32 slot_parity;
  	u32 blkID;
} trk_tab_item_t;

typedef struct {
	baddr_t BufStart; 
	baddr_t BufEnd;
	char Status;
} wait_req_t; 


typedef struct {
	wait_req_t wait_req[9]; 
	u32 PreRead_SlotNo;	
} preRead_info_t; 





/* hammer 1120 */
typedef struct {
  u32 blk_ID;
  u32 blk_ADR;
  u32 next_blkID;
  u32 parity;
} buf_indicator;

typedef struct {
  u32 device_ID;
  u32 reserved;  
  u32 buf_head;
  baddr_t buf_start;
  baddr_t buf_end;
  u32 buf_size;
  u32 buf_cnt;
  u32 parity;
} pbm_slot;

/* 21/11/2001 added by hammer for pend request that PBM to Engine When Engine can't accept new coming request */
typedef struct {    
    buf_req_t   *req_head;
    buf_req_t   *req_tail;
    unsigned int      count;
} p_queue_t;


/* add by hammer for random performance */
typedef struct _slot_node {
    U32 prev_SlotNo;
    U32 next_SlotNo;
    U32 NodeTrue;
    baddr_t     stripe_start;
    
} slot_node;

typedef struct _slot_queue_t{
	slot_node slot_tag[BUF_MAX_TRK_ITEM];
	U32 SlotHead;
	U32 SlotEnd;
	U32 SlotCurrent;
	U32 SlotActive;
	U32 SlotCount;
	U32 SortDir;
	} slot_queue_t;
/* end of random performance enhance */



typedef struct {
    trk_tab_item_t  items[BUF_MAX_TRK_ITEM];
    buf_req_t   *req_head;
    buf_req_t   *req_tail;
    unsigned int      count;
} trk_tab_t;

extern int buf_flush(void);
extern void buf_init(unsigned long availMemSize, unsigned long availMemSize1);
extern void buf_submit(REQ *req);
extern void Engine_RAID_Return(REQ *req);
extern int buf_return(REQ* req);

extern void mm_memmove( void* pdst,void* psrc, unsigned long cnt);
extern int mm_memcmp( unsigned char* p1,unsigned char* p2, unsigned long cnt );
/* Modified by crimson. 11/19, Y2001*/
extern void mm_memset(void* pmem, char b, unsigned long cnt);
extern void mm_memtest(void* pmem, unsigned long cnt);
#ifdef NT
#pragma pack()
#endif



#define BLK_PER_STRIPE_BLK(dev_id)  aryDef[dev_id].hStripeBlockSize
#define N_BLK_DEVS(dev_id)          aryDef[dev_id].bLogicalWidth
//modified by isaac according hammer's advice
#define N_STRIPE(dev_id)            (aryDef[dev_id].wSize/aryDef[dev_id].hStripeBlockSize)
#define RAID_TYPE(dev_id)			aryDef[dev_id].blRAID

/*extern unsigned long buf_get_stripe_size(int dev_id);*/

/*#define buf_get_stripe_size(dev_id) (BLK_PER_STRIPE_BLK * (N_BLK_DEVS))*/

#define buf_get_stripe_size(dev_id) (aryStripeSize[(dev_id)])

/* /== for RAID 5
/#define buf_get_stripe_size(dev_id) (BLK_PER_STRIPE_BLK * (N_BLK_DEVS-1))*/

#define buf_get_stripe_blk_right_boundary(dev_id, blkno) (howmany(blkno + 1, BLK_PER_STRIPE_BLK(dev_id)) * BLK_PER_STRIPE_BLK(dev_id) - 1)
#define buf_get_stripe_blk_left_boundary(dev_id, blkno) ((howmany(blkno + 1, BLK_PER_STRIPE_BLK(dev_id)) - 1) * BLK_PER_STRIPE_BLK(dev_id))
#define buf_get_stripe_left_boundary(dev_id, blkno) ((howmany(blkno + 1, buf_get_stripe_size(dev_id)) - 1) * buf_get_stripe_size(dev_id))
#define buf_get_stripe_right_boundary(dev_id, blkno) (howmany(blkno + 1, buf_get_stripe_size(dev_id)) * buf_get_stripe_size(dev_id) - 1)
#define buf_get_stripe_blk_size(dev_id) BLK_PER_STRIPE_BLK(dev_id)
#define buf_can_pre_read(dev_id) 0
#define buf_get_max_blk(dev_id) (BLK_PER_STRIPE_BLK(dev_id) * N_STRIPE(dev_id) - 1)
#define buf_blk2stripe(dev_id, blkno) (howmany(blkno + 1, buf_get_stripe_size(dev_id)) - 1)



extern void DisableInt(void);
extern void EnableInt(void);

#define ECS DisableInt()
#define LCS EnableInt()


#define GET_K                   1

#define IS_SCSI_RW(req) ((req->bCDB[0] == scsiREAD_6) ||(req->bCDB[0] == scsiREAD_10) ||(req->bCDB[0] == scsiWRITE_6) ||(req->bCDB[0] == scsiWRITE_10))

#define IS_SCSI_R(req)	((req->bCDB[0] == scsiREAD_6) ||(req->bCDB[0] == scsiREAD_10))

#define IS_SCSI_W(req)	((req->bCDB[0] == scsiWRITE_6) ||(req->bCDB[0] == scsiWRITE_10))

//#define ARRAYID2CARD(dev_id) (dev_id<4?0:1)


#endif /* __BUF_H__ */
