/***************************************************************************
 *
 * Copyright (c) 1995 Promise Technology, Inc.  All rights reserved.
 * No part of this document may be reproduced or transmitted in any form or
 * by any means, electronic or mechanical, for any purpose, without the
 * express written permission of Promise Technology, Inc.
 *
 *
 * scsi-3.h - contains scsi-3 interface definition and mappings
 *
 ***************************************************************************/


/*
 * scsi3 opcode and CDB bits
 */

#define svucSCSI_COMMAND                0x60

#define scsiINQUIRY                     0x12
  #define blINQUIRY_EVPD                bit(0)
  #define bDISK_UNIT_INFO_PAGE          0xC0
  #define DISK_UNIT_INFO_PAGE_SIZE      0xF0
  #define STANDARD_INQUIRY_PAGE_SIZE    0x30
  #define INQUIRY_PRODUCT_NAME_OFFSET   0x10
#define scsiMODE_SELECT_6               0x15
#define scsiMODE_SELECT_10              0x55
#define scsiMODE_SENSE_6                0x1A
#define scsiMODE_SENSE_10               0x5A
  #define bCURRENT_PARAMETER            0x00
  #define bCHANGEABLE_PARAMETER         0x01
  #define bDEFAULT_PARAMETER            0x02
  #define bSAVED_PARAMETER              0x03
  #define bARRAY_MAPPING_MODE_PAGE      0x20
  #define bCONTROLLER_SPECIFIC_PAGE     0x21
  #define bMAINTENANCE_CONTROL_PAGE     0x22
  #define bALL_PAGES                    0x3F
#define scsiREAD_6                      0x08
#define scsiREAD_10                     0x28
#define scsiREAD_CAPACITY               0x25
#define scsiREQUEST_SENSE               0x03
  #define REQUEST_SENSE_DATA_SIZE       0x20
#define scsiREZERO_UNIT                 0x01
#define scsiSEEK_6                      0x0B
#define scsiSEEK_10                     0x2B
#define scsiSTART_STOP_UNIT             0x1B
#define scsiTEST_UNIT_READY             0x00
#define scsiVERIFY                      0x2F
#define scsiWRITE_6                     0x0A
#define scsiWRITE_10                    0x2A
#define scsiWRITE_VERIFY                0x2E
#define scsiFLUSH                                               0x35                    
#define scsiSYNCHRONIZE_CACHE           0x35

/*
 * SCSI 3 vendor unique opcodes
 */

#define svucARRAY_COMMAND               0xE0
  #define acmdMASK                      0x1F
  #define acmdREPORT_DEVICE_MAP         0x00
  #define acmdSAVE_CONFIG               0x01
  #define acmdREAD_PHYSICAL             0x02
  #define acmdWRITE_PHYSICAL            0x03
  #define acmdSCAN_CONFIG               0x04
  #define acmdFORCE_DISK_DOWN           0x05
  #define acmdFORCE_DISK_UP             0x06
  #define acmdFORCE_DISK_SPARE          0x07
  #define acmdREBUILD_DISK              0x08
  #define acmdCHECK_INTEGRITY           0x09
  #define acmdSYNC_REDUNDANCY           0x0A
  #define acmdMANUAL_THROTTLE           0x0B
  #define acmdHEART_BEAT                0x0C
  #define acmdCOMMAND                   0x0D    /* total number of 0xE0 cmds */
  #define acmdSMART_COMMAND             0x0E
  #define  acmdCONV_EXP                 0x0F    /* Add by David Wei*/
#define svucCONTROLLER_COMMAND          0xE1
  #define ccmdCOMMAND                   0x02
  #define ccmdPING                      0x00
  #define ccmdDEACTIVATE                0x01


/*
 * scsi status or bTargetStatus in REQ Structure
 */

#define ssGOOD                          0x00    /* no error */
#define ssCHECK_CONDITION               0x02    /* check sense data for status */
#define ssCONDITION_MET                 0x04    /* should not happen */
#define ssBUSY                          0x08    /* RAID Engine is busy, retry after some time */
#define ssINTERMEDIATE_GOOD             0x10    /* should not happen */
#define ssINTERMEDIATE_MET              0x14    /* should not happen */
#define ssRESERVATION_CONFLICT          0x18    /* should not happen */
#define ssCOMMAND_TERMINATED            0x22    /* should not happen */
#define ssQUEUE_FULL                    0x28    /* RAID Engine is busy, retry after some time */
#define ssLOCK                          0x80 

/*
 * scsi sense key, ASC, and ASCQ
 */

#define skNO_SENSE                      0x00    /* no additional status */
#define skRECOVERED_ERROR               0x01    /* cmd completed using recovery */
#define skNOT_READY                     0x02    /* array is busy, retry after some time */
#define skMEDIUM_ERROR                  0x03    /* auto-reassign/recovery failed */
#define skHARDWARE_ERROR                0x04    /* disk hardware error */
#define skILLEGAL_REQUEST               0x05    /* invalid bits in CDB */
#define skUNIT_ATTENTION                0x06    /* special event just happened */
#define skDATA_PROTECT                  0x07    /* should not happen */
#define skBLANK_CHECK                   0x08    /* should not happen */
#define skCOPY_ABORTED                  0x0A    /* should not happen */
#define skABORTED_COMMAND               0x0B    /* should not happen */
#define skEQUAL                         0x0C    /* should not happen */
#define skVOLUME_OVERFLOW               0x0D    /* should not happen */
#define skMISCOMPARE                    0x0E    /* should not happen */

#define saNO_ASC                        0x00
  #define sqNO_ASCQ                       0x00
#define saLUN_NOT_READY                 0x04
  #define sqINTERVENTION_REQUIRED         0x03
  #define sqMAINTENANCE_IN_PROGRESS       0x80
#define saECC_ERROR                     0x10
#define saID_ADDR_MARK_NOT_FOUND        0x12
#define saINVALID_OPCODE                0x20
#define saLBA_OUT_OF_RANGE              0x21
#define saINVALID_FEILD_IN_CDB          0x24
#define saLOGICAL_UNIT_NOT_SUPPORTED    0x25
#define saINVALID_FIELD_IN_PARAMETER    0x26
#define saINTERNAL_TARGET_FAILURE       0x44
