/***************************************************************************
 *
 * Copyright (c) 1995 Promise Technology, Inc.  All rights reserved.
 * No part of this document may be reproduced or transmitted in any form or
 * by any means, electronic or mechanical, for any purpose, without the
 * express written permission of Promise Technology, Inc.
 *
 *
 * struct.h - defines data structures for controller, LUN, disk...
 *              S.Zhu      07/21/97     Check integrity for NT
 *
 ***************************************************************************/

/*#ifdef NT_MSC*/
#ifdef NT
#pragma pack (1)        /*12/16/97 S.Zhu*/
#endif
/*#endif*/

#define _PBM_    1
/*#define _INVERT_  1*/

typedef  void* PVOID;
/*typedef  unsigned char U8;*/

#if defined (__LINUX__) | (__FreeBSD__)				/* 	Linus Chen 	*/
    #define PACKED __attribute__((packed))
#else  
    #define PACKED
#endif

#define DEV_STATUS_OK               (0x00000000)
#define DEV_STATUS_WARNING          (0x80000001)
#define DEV_STATUS_ERROR            (0x80000002)


/* array status */
#define ARY_STATUS_OK                   (DEV_STATUS_OK)
#define ARY_STATUS_WARNING              (DEV_STATUS_WARNING)
#define ARY_STATUS_ERROR                (DEV_STATUS_ERROR)
#define ARY_STATUS_CHECKING_INTEGRITY   (0x00000100)
#define ARY_STATUS_REBUILDING           (0x00000200)
#define ARY_STATUS_EXPANDING            (0x00000800)
#define ARY_STATUS_REBUILD_PAUSE         (0x00000400)  /* Add by David Wei means stop rebuild or conversion expansion*/
#define ARY_STATUS_CONVEXP_PAUSE         (0x00001000)
#define ARY_STATUS_SYNC_PAUSE            (0x00002000)
//added by isaac for raid5 initialization
#define ARY_STATUS_RAID5_INITIALIZATION  (0X00004000)

#define DSK_STATUS_OK               (DEV_STATUS_OK)
#define DSK_STATUS_WARNING          (DEV_STATUS_WARNING)
#define DSK_STATUS_ERROR            (DEV_STATUS_ERROR)
#define DSK_STATUS_FREE             (0x00000100)
#define DSK_STATUS_ASSIGNED         (0x00000200)
#define DSK_STATUS_SPARE            (0x00000300)
#define DSK_STATUS_CLAIMED          (DSK_STATUS_ASSIGNED)


#define MAX_CONV_EXP_RATE                           0x256
#define MAX_SYNC_RATE                               0x256
#define MAX_REBUILD_RATE							0x256

/***************************************************************************
 *
 * Disk Configuration Structures
 *
 ***************************************************************************/



 
/*
 * disk access characteristics
 */
#ifdef CACHE
#define RAID5_NUM_OF_BUFFER             1
#else
#define RAID5_NUM_OF_BUFFER             16  /* It was 8*/
#endif

#define blDISK_IS_PRESENT_HERE          bit(0)  /* blDiskFlag */
#define blDISK_HAS_NO_SIGNATURE         bit(1)
//Label_bp begin
#define blDISK_DOES_NOT_DMA             bit(2)
#define blDISK_DOWN_BY_ERROR            bit(2)
//Label_bp end
#define blNEW_DRIVE                     bit(3)  /* 12/16/97 Q.W  */
#define blDISABLESORT                   bit(4)
#define blDELETE_DISK                   bit(5)    
#define blRESET_IN_PROGRESS             bit(6)  /* id 0 only */
#define blCHANNEL_BUSY                  bit(7)  /* id 0 only */

#define TICK_INIT                       0

/*
 * disk info
 */

#define blDISK_EXISTS                   bit(0)  /* blFlag */

#define MODEL_NAME_SIZE                 20

typedef struct disk_info {
  ByTE blFlag;
  ByTE bCh;
  ByTE bID;
  ByTE bReserved;
  WORD wSize;                                   /* in number of sectors */
  ByTE bModel[MODEL_NAME_SIZE];
} PACKED DISK_INFO, PACKED buf_space *bpDISK_INFO, PACKED app_space *apDISK_INFO;


/***************************************************************************
 *
 * RAID Configuration Structures
 *
 ***************************************************************************/

#define RAID_DEF_FORMAT                 (0x00020000)
  #define DEF_FORMAT_MINOR              (0x0000FFFF)
  #define DEF_MINOR_SHIFT               0x10
  #define MAX_ARRAY_SUPPORTED           0x08
  #define MAX_ADATPER_SUPPORTED         0x02  //added by isaac for dualcard
  #define ACTUAL_ARRAY_SUPPORTED        0x08  //modified by isaac from 4 for dualcard
  #define MAX_ARRAY_PER_CARD             0x04  //added by isaac for dualcard
  #define MAX_CHANNEL_PER_CARD           0x04  //added by isaac for dualcard  
#define MAX_SEGMENT_PER_ARRAY         0x02
  #define MAX_STREAM_PER_ARRAY          0x08
#define MAX_STREAM_PER_ARRAY1          0x08
//Label_bp begin
#define MAX_BLOCKS                      (MAX_CHANNEL_SUPPORTED*MAX_ID_PER_CHANNEL*MAX_BLOCK_PER_DISK)
//Label_bp end
#define MAX_BLOCK_PER_STREAM          0x01
  #define MAX_BLOCK_PER_DISK            0x01
  #define SERIAL_SIZE                   0x08

  #define SECTOR_SIZE                   0x0200
  #define SECTOR_SHIFT                  0x09
  #define SECTOR_SIZE_MASK              0x01FF
//modified by isaac for dualcard
#define MAX_CHANNEL_SUPPORTED           0x08 /*0x08   ==it was 8 modified 9.25 for raidapi*/
#define MAX_ID_PER_CHANNEL              0x02	/*== it was 2*/
#define MAX_BLOCK_PER_CARD              (MAX_CHANNEL_SUPPORTED*MAX_ID_PER_CHANNEL*MAX_BLOCK_PER_DISK)
//#define MAX_SEQUENCE_PER_CMD            0x0a*32 //0x14   /* Modifed by David Wei 2001 1227 for stripe size==4k*/
#define MAX_SEQUENCE_PER_CMD            0x14   /* Modifed by David Wei 2001 1227 for stripe size==4k*/


#define MY_MAGIC                        0x5A

#define ERRORLOG_NUM                    20

/* modified by crimson 05/27. */
//#define wMIN_VIR_BUFFER_SIZE            sizeof(SBUF )*(4)+sizeof(BBUF )*(1)+32
#define wMIN_VIR_BUFFER_SIZE            sizeof(BBUF )*(3)

/* 
    hammer 1120 
    Added by crimson, 12/12, Y2001 
    According to hammer's new pbm algorithm
*/
#define PBM_MAX_BUFHEAD 2048
#define PBM_MIN_BUFHEAD 64

#define PBM_REQ_BUF_SIZE		((sizeof(REQ)+4)/4)*4*(360			)
// hammer 0221
#define PBM_BUFREQ_BUF_SIZE		((sizeof(buf_req_t)+4)/4)*4*(1024/*150 */		)
/* #define PBM_BUFHEAD_BUF_SIZE	((sizeof(buf_head_t)+4)/4)*4*(1024	) */
#define PBM_BUFHEAD_BUF_SIZE	((sizeof(buf_head_t)+4)/4)*4*(PBM_MAX_BUFHEAD	)

// Hammer 2002 0520 reduce the resource requirement of PBM
#define MIN_PBM_REQ_BUF_SIZE		((sizeof(REQ)+4)/4)*4*(8)
// hammer 0221
#define MIN_PBM_BUFREQ_BUF_SIZE		((sizeof(buf_req_t)+4)/4)*4*(64)
/* #define PBM_BUFHEAD_BUF_SIZE	((sizeof(buf_head_t)+4)/4)*4*(1024	) */
#define MIN_PBM_BUFHEAD_BUF_SIZE	((sizeof(buf_head_t)+4)/4)*4*(PBM_MIN_BUFHEAD	)

/*==#define PBM_SG_BUF_SIZE			((sizeof(SG)+4)/4)*4*(100			)*/
/*==#define PBM_SECTOR_BUF_SIZE		512*(81920			)*/

/*
	09/21, Y2001.
	Modified by crimson for the size of READBUF_CONTROL_BLOCK, PARITYBUF_CONTROL_BLOCK and RECOVERBUF_CONTROL_BLOCK
	
	
*/
//#define wVIR_BUFFER_SIZE                sizeof(SBUF )*(2+MAX_CONCURRENT_THREAD)+sizeof(BBUF )*(6)+sizeof(DTASK)*(8*MAX_CONCURRENT_THREAD)+((sizeof(XORBUF)+4)/4)*4*(2+MAX_CONCURRENT_THREAD)+sizeof(READBUF_CONTROL_BLOCK)*(MAX_CONCURRENT_THREAD)+sizeof(PARITYBUF_CONTROL_BLOCK)*(MAX_CONCURRENT_THREAD)+sizeof(RECOVERBUF_CONTROL_BLOCK)*(MAX_CONCURRENT_THREAD)+sizeof(CHPRDTBL)*(8*MAX_CONCURRENT_THREAD)+((sizeof(REQ)+4)/4)*4*(6)+((sizeof(EVENT)+4)/4)*4*(8)+sizeof(ERLOG)*(ERRORLOG_NUM            )+sizeof(REBUILDBUF_CONTROL_BLOCK) +32 +	(PBM_REQ_BUF_SIZE)+  (PBM_BUFREQ_BUF_SIZE)+ (PBM_BUFHEAD_BUF_SIZE)
#ifdef NT
#define wVIR_BUFFER_SIZE    sizeof(SBUF )*(2+MAX_CONCURRENT_THREAD)+\
              sizeof(BBUF )*(6)+sizeof(DTASK)*(8*MAX_CONCURRENT_THREAD)+\
	      ((sizeof(XORBUF)+4)/4)*4*(2+MAX_CONCURRENT_THREAD)+\
	      sizeof(READBUF_CONTROL_BLOCK)*(MAX_CONCURRENT_THREAD)*2+\
	      sizeof(PARITYBUF_CONTROL_BLOCK)*(MAX_CONCURRENT_THREAD)*2+\
	      sizeof(RECOVERBUF_CONTROL_BLOCK)*(MAX_CONCURRENT_THREAD)*2+\                    sizeof(CHPRDTBL)*(8*MAX_CONCURRENT_THREAD)+\
	      ((sizeof(REQ)+4)/4)*4*(6)+\
	      ((sizeof(EVENT)+4)/4)*4*(8)+\
	      sizeof(ERLOG)*(ERRORLOG_NUM)+\
	      sizeof(REBUILDBUF_CONTROL_BLOCK)*2+\
	      32
#else
//event resource count modified by isaac	      
#define wVIR_BUFFER_SIZE    sizeof(SBUF )*(2+MAX_CONCURRENT_THREAD)+sizeof(BBUF )*(6)+sizeof(DTASK)*(8*MAX_CONCURRENT_THREAD)+((sizeof(XORBUF)+4)/4)*4*(2+MAX_CONCURRENT_THREAD)+sizeof(READBUF_CONTROL_BLOCK)*(MAX_CONCURRENT_THREAD)*2+sizeof(PARITYBUF_CONTROL_BLOCK)*(MAX_CONCURRENT_THREAD)*2+sizeof(RECOVERBUF_CONTROL_BLOCK)*(MAX_CONCURRENT_THREAD)*2+((sizeof(REQ)+4)/4)*4*(6)+((sizeof(EVENT)+4)/4)*4*(16)+sizeof(ERLOG)*(ERRORLOG_NUM)+sizeof(REBUILDBUF_CONTROL_BLOCK)*2+32
#endif                                        
#ifndef _PBM_
#define wPHY_BUFFER_SIZE                sizeof(SBUF )*(2+MAX_CONCURRENT_THREAD)+\
					                    sizeof(BBUF )*(6                      )+\
					                    sizeof(DTASK)*(8*MAX_CONCURRENT_THREAD)+\
					                    ((sizeof(XORBUF)+4)/4)*4*(2+MAX_CONCURRENT_THREAD)+\
                                        sizeof(READBUF)*8*(RAID5_NUM_OF_BUFFER+1)+\
                                        sizeof(PARITYBUF)*4*(RAID5_NUM_OF_BUFFER+1)+\
                                        sizeof(RECOVERBUF)*8*(RAID5_NUM_OF_BUFFER+1)+\
                                        sizeof(CHPRDTBL)*(8*MAX_CONCURRENT_THREAD)+\
                                        ((sizeof(REQ)+4)/4)*4*(6               )+\
                                        ((sizeof(EVENT)+4)/4)*4*(8             )+\
                                         sizeof(ERLOG)*(ERRORLOG_NUM            )+\
                                        (REBUILDBUF_DATA_BLOCK)+\
                                        32
#else


#define wPHY_BUFFER_SIZE                sizeof(READBUF_DATA_BLOCK)*(MAX_CONCURRENT_THREAD  * MAX_CHANNEL_SUPPORTED/2)+sizeof(PARITYBUF_DATA_BLOCK)*(MAX_CONCURRENT_THREAD * MAX_CHANNEL_SUPPORTED/2)+sizeof(RECOVERBUF_DATA_BLOCK)*(MAX_CONCURRENT_THREAD  * MAX_CHANNEL_SUPPORTED/2)+sizeof(REBUILDBUF_DATA_BLOCK)+32	

#endif


#define wMIN_NON_PAGED_BUFFER_SIZE      sizeof(SBUF )*(2+MIN_CONCURRENT_THREAD)+sizeof(DTASK)*(8*MIN_CONCURRENT_THREAD)+((sizeof(XORBUF)+4)/4)*4*(2+MIN_CONCURRENT_THREAD)+sizeof(READBUF_CONTROL_BLOCK)*(MIN_CONCURRENT_THREAD)+sizeof(PARITYBUF_CONTROL_BLOCK)*(MIN_CONCURRENT_THREAD)+sizeof(RECOVERBUF_CONTROL_BLOCK)*(MIN_CONCURRENT_THREAD)+sizeof(CHPRDTBL)*(8*MIN_CONCURRENT_THREAD)+((sizeof(REQ)+4)/4)*4*(3)+((sizeof(EVENT)+4)/4)*4*(4)+sizeof(ERLOG)*(4)+sizeof(REBUILDBUF_CONTROL_BLOCK) +32

#define wVIR_PBM_BUFFER_SIZE   (PBM_REQ_BUF_SIZE)+  (PBM_BUFREQ_BUF_SIZE)+ (PBM_BUFHEAD_BUF_SIZE)
// Hammer 2002 0520 reduce the resource requirement of PBM
#define wMIN_VIR_PBM_BUFFER_SIZE		(MIN_PBM_REQ_BUF_SIZE)+  (MIN_PBM_BUFREQ_BUF_SIZE)+ (MIN_PBM_BUFHEAD_BUF_SIZE)

#define wMIN_PHY_BUFFER_SIZE            wPHY_BUFFER_SIZE   

/*added by crimson.06/07/01*/
#define  wREAL_PHY_BUFFER_SIZE		0x00

/*
 * BLK describes where does a section of sectors in an array comes from?
 * first 2 byte must sync with STRM.
 * byte 2~11 must sync with STREAM
 * see TranslateLUN_ToArray().
 */

#define blVALID_BLK                     bit(0)  /* blFlag */
#define blWORKING_BLK                   bit(1)
#define blASSIGNED_BLK                  bit(2)
#define blSPARE_BLK                     bit(3)
#define blDUPLICATE_BLK                 bit(4)
#define blREDIRECT_BLK                  bit(5)
#define blMARK_FOR_DOWN                 bit(6)
#define blNEED_HOT_SPARE                bit(6)
#define blMARK_FOR_UP                   bit(7)

typedef struct block {
  BoOL blFlag;
  ByTE bStream;
  ByTE bCh;
  ByTE bID;
  ByTE bSerial[SERIAL_SIZE];
  WORD wStartLBA;
  WORD wSize;
  WORD wWatermark;
} PACKED BLOCK, PACKED buf_space *bpBLOCK, PACKED dat_space *dpBLOCK;


/*
 * STREAM describes how a stream is formed using which disk and that disk's
 * status.
 * first 2 byte must sync with STRM.
 * byte 2~11 must sync with BLOCK
 */

typedef struct stream {
  BoOL blFlag;                                  /* see BLOCK.blFlag */
  ByTE bIndex;
  ByTE bCh;
  ByTE bID;
  ByTE bSerial[SERIAL_SIZE];
} PACKED STREAM, PACKED buf_space *bpSTREAM;


/*
 * defines an array using STREAM structure.  also specify properties of this
 * array.  first 20 byte of this structure must sync with ARY,
 * strm must be last field, see p_lunConsolidateLUN()
 * see TranslateLUN_ToArray().
 * also location of every byte here is tied to ConsolidateLUN().
 */

#define blVALID_LUN                     bit(0)  /* blFlag */
#define blWORKING_LUN                   bit(1)
#define blINITIALIZED_LUN               bit(2)
#define blARY_READY                     bit(3)
#define blRECOVERING_LUN                bit(4)
#define blMARKED_LUN                    bit(5)
#define blCONV_EXP_LUN                  bit(6)
#define blBOUNDARY						bit(7)


#define blMIRROR                        bit(0)  /* blRAID */
#define blPARITY                        bit(1)
#define blDISTRIBUTED                   bit(2)
#define blSPAN                          bit(3)
#define blJBOD                          bit(4)  //added by isaac for JBOD
#define blRAID0                         0 
#define blRAID1                         blMIRROR
#define blRAID4                         blPARITY
#define blRAID5                         blDISTRIBUTED

#define ARRAY_MODE_STRIPPING            (blRAID0)
#define ARRAY_MODE_MIRRORING            (blRAID1)
#define ARRAY_MODE_STRIPMIRRORING       (blRAID1|0x100)
#define ARRAY_MODE_PARITY               (blRAID4)
#define ARRAY_MODE_PARITY_DISTRIBUTED   (blRAID5)
#define ARRAY_MODE_SPAN                 (blSPAN)
#define ARRAY_MODE_JBOD                 (blJBOD)  //added by isaac for JBOD
#define ARRAY_MODE_UNKNOWN              (0xFFFFFFFF)


/*   RAID5 status  */
#define blREADBUF                       bit(0)
#define blREADRECOVERY                  bit(1)
#define blWRITERECOVERY                 bit(2)
#define blReadPartOfBlock               bit(3)
#define blReadPartOfParity              bit(4)
#define blReadTailPartOfBlock           bit(5)
#define blFirstXorRead                  bit(6)
#define blSecondXorRead                 bit(7)


/* RMW Status */
#define blRMW                                                   bit(0)
#define blONEDRIVEBLOCK                                 bit(1)  
#define blRMWLINE2                                              bit(2)
#define blSPECIALSECOND                                 bit(3)
#define blSPECIALFINAL                                  bit(4)
#define blPARITYONLY                    bit(5)  
#define blSPECIALRMW                    bit(6)
#define blNORMALRMW                     bit(7)


typedef struct lun {
  HALF hGeneration;
  BoOL blFlag;
  BoOL blRAID;
  ByTE bWidth;
  ByTE bStripeBlockPower;                       /* power of 2 */
  ByTE bLogicalWidth;
  ByTE bID;
  WORD wSize;                                   /* # of sectors */
  HALF hCyl;                                    /* # of cyl */
  ByTE bHd;                                     /* # of hd per cyl */
  ByTE bSctr;                                   /* # of sctr per hd */
  ByTE bSerial[SERIAL_SIZE];
 
  STREAM strm[MAX_STREAM_PER_ARRAY];
  ByTE Name[64];
  

   ByTE bInit;
   
   /* 
   	Added by crimson. 11/19, Y2001 
   	For DWORD alignment 
   */
   ByTE Reserved1[7];

  /*
   Added by crimson. 10/23, Y2001 .
   For Get wWaterMark from SRAM in conversion
   
   removed by crimson. 11/19, Y2001
   We adopt another scheme.
  */
  /* WORD wWatermark; */
} PACKED LUN, PACKED buf_space *bpLUN;


#define SIGNATURE_SIZE                  24

typedef struct id {
  ByTE bSignature[SIGNATURE_SIZE];              /* Promise Technology, Inc. */
  WORD wFormat;
  ByTE bSerial[SERIAL_SIZE];
  ByTE bDate[6];                                /* 2 byte days + 4 byte ticks */
  ByTE bReserved[6];
} PACKED ID, PACKED buf_space *bpID;


#define blVALID_ASSOCIATOR              bit(7)  /* blFlag */

typedef struct assoc {
  ByTE  blFlag;
  ByTE  bReserved[3];
  BLOCK blk;
  LUN   lun;
    /* 
    Added by crimson. 10/23, Y2001
    
    Removed by crimson. 11/19, Y2001 
    We adopt another scheme.
    */
  /* LUN   lun1; */
} PACKED ASSOC, PACKED buf_space *bpASSOC;


/*
 * reserved sector structure
 */

#define NUMBER_OF_RESERVED_SECTOR       4 //modified by isaac for dualcard, original is 4       /* has checksum protection */
  #define bSIGNATURE_PAGE               0
  #define bDEFINITION_PAGE              1
  #define bCHECKSUM_PAGE                3
#define bCONTROLLER_PAGE                4

typedef struct disk {
  ID    id;                                     /* sector 0 */
  ByTE  bRestOfSector0[512-sizeof(ID)];
  ASSOC links[MAX_BLOCK_PER_DISK];              /* sector 1 */
  ByTE  bRestOfSector1[512-sizeof(ASSOC)];
  ByTE  bRestOfSector2[512];                    /* sector 2 */
  ByTE  bRestOfSector3[512-4];                  /* sector 3 */
  WORD  wChecksum;
} PACKED DISK, PACKED buf_space *bpDISK;


/***************************************************************************
 *
 * Operating Structures: structures that are optimized for stripping use
 *
 ***************************************************************************/

/*
 * first 2 bytes of this structure must sync with STREAM.
 * see TranslateLUN_ToArray().
 */

typedef struct strm {
  BoOL blFlag;                                  /* see BLOCK.blFlag */
  ByTE bIndex;
  ByTE bDuplicateIndex;
  ByTE bRedirectIndex;
} PACKED STRM, PACKED dat_space *dpSTRM;


/*
 * first 20 bytes this structure must sync with LUN.
 * see TranslateLUN_ToArray().
 */

typedef struct array {
  HALF hGeneration;
  BoOL blFlag;
  BoOL blRAID;
  ByTE bWidth;
  ByTE bStripeBlockPower;                       /* in power of 2 sctrs */
  ByTE bLogicalWidth;
  ByTE bID;
  WORD wSize;                                   /* # of sectors */
  HALF hCyl;                                    /* # of cyl */
  ByTE bHd;                                     /* # of hd per cyl */
  ByTE bSctr;                                   /* # of sctr per hd */
  ByTE bSerial[SERIAL_SIZE];
  WORD wWatermark;
  HALF hStripeBlockSize;                        /* in sctrs */
  HALF hStripeSize;                             /* in sctrs */
  STRM strm[MAX_STREAM_PER_ARRAY];              /* index into block database */
  ByTE Name[64]; 
  ByTE bInit;
   /* 
   	Added by crimson. 11/19, Y2001 
   	For DWORD alignment 
   */
   ByTE Reserved1[7];
} PACKED ARY, PACKED dat_space *dpARY;


/*
typedef struct array1{
    U8 blSourceRAID;
    U8 blTargetRAID;
    U8  bSourceLogicWidth;          
    U8  bTargetLogicWidth;
    U8  bSourceWidth;
    U8  bTargetWidth;
    U16 hSourceStripeSize;
    U16 hTargetStripeSize;
    U32 wConvWatermark;
    U32 wTargetSize;  
}ARY1,dat_space *dpARY1;
*/


typedef struct q {                              /* sync with SBUF and BBUF */
  struct q str_space *p_next;
  ByTE   bReserved[3];
  ByTE   bMagic;
  ByTE   bReserved1[4];
  struct q buf_space *p_link;
  U32    buf;                                   /* added by crimson 05/27 - used for nonaccessable mem */
} PACKED Q, PACKED buf_space *bpQ, PACKED str_space *spQ;




/*#define BIG_SCRATCH_BUFFER_SIZE         0x800*/

#ifdef BIOS
#define BIG_SCRATCH_BUFFER_SIZE         0x800
#else
#define BIG_SCRATCH_BUFFER_SIZE         0x1000  /*modified by Crimson 2001/09/24*/
#endif


/*#define XOR_PRD_BUFFER_SIZE               0x100*/

#ifdef CACHE
#define PARITY_BUFFER_SIZE                0x8000
#define READ_XOR_BUFFER_SIZE              0x10                
#else
#define PARITY_BUFFER_SIZE                0x10000
#define READ_XOR_BUFFER_SIZE              0x10000              
#endif

#define READSG_BUFFER_SIZE                0x200      
#define PARITYSG_BUFFER_SIZE              0x200

#define RECOVER_BUFFER_SIZE               0x10000      
#define RECOVERSG_BUFFER_SIZE             0x200          
/* Modified by crimson. 06/08/01. For Test only.*/
#define REBUILD_BUFFER_SIZE               0x40000
/*#define REBUILD_BUFFER_SIZE               0x4000*/
#define REBUILDSG_BUFFER_SIZE             0x100
#define PRD_TBL_SIZE                      0x100  

typedef struct bbuf {
  spQ  p_next;                                  /* link pointer */
  ByTE bReserved[8];
  union {
    bpQ  p_link;
    WORD w;
  } link;
  ByTE buf[BIG_SCRATCH_BUFFER_SIZE];
} PACKED BBUF, PACKED buf_space *bpBBUF;


/*#define REBUILD_BUFFER_SIZE        0x8000*/
/*
typedef struct xbuf {
    spQ  p_next;                                  
    ByTE bReserved[8];
    union {
	bpQ  p_link;
	WORD w;
    } link;
    ByTE buf[REBUILD_BUFFER_SIZE];
} XBUF, buf_space *bpXBUF;
*/
#define blLAST_PRD_ENTRY                bit(7)  /* blFlag */
#define blNON_OS                  bit(0)

/*
 * describes an entry in PRD table, note the structure is in little Endian.
 */


typedef struct prd {
    WORD wStartAddr;                              /* physical addr */
    HALF hCount;                                  /* 16 bit count */
    ByTE bReserved;
    BoOL blFlag;
} PACKED PRD, PACKED buf_space *bpPRD;

/***************************************************************************
 *
 * Command packet and structures:
 *
 ***************************************************************************/

/*
 * a group of tasks to be executed concurrently is called a sequence.  all tasks
 * within a sequence must be completed before the next sequence is started.
 */

#define blERROR_IN_SEQUENCE             bit(7)  /* blFlag */
#define blHardwareError             bit(3)
#define blCompareNotSame            bit(2)
#define blCompareSame               bit(1)
#define blIsCompare                 bit(0) 
 

#define TASK_SEQ                        0x00    /* bType */
#define STATUS_SEQ                      0x01
#define TAGGING_SEQ                     0x02
#define CONTINUE_IF_IDLE                0x03
#define SIGNAL_SEQ                      0x04
#define RELEASE_INTERNAL_SEQ            0x05
#define ARITHMETIC_SEQ                  0x07
#define ASSIGN_SEQ                      0x08
#define ASSIGN_BYTE_SEQ                 0x09
#define EXECUTE_SEQ                     0x0A
#define EXECUTE_SEQ_SEQ                 0x0B
#define FOR_SEQ                         0x0C
#define GOTO_SEQ                        0x0D
#define FINISH_SEQ                      0x0E       
#define XOR_SEQ                         0x0F
#define CONTINUE_CONV_IF_IDLE           0x10
#define CONV_READ_CONTINUE_IF_IDLE      0x11
#define COMPARE_SEQ			0x12

typedef struct seq {
  union {
    spQ     p_task;
    struct  req buf_space *p_rq;
    struct  seq buf_space *p_sq;
  } branch;
  BoOL blFlag;
  ByTE bType;
  ByTE bCount;
  ByTE bMagic;                                  /* signature byte to id struct */
  union {
    spQ     p_task;
    dpwBUF  p_w;
    dpBUF   p_b;
    BoOL    (* function)(void *);
  } action;
  union {
    struct  req str_space *p_rq;                /* -> assoc. RAID cmd struct */
    WORD    w;
    ByTE    b;
    struct event buf_space *p_ev;
  } param;
  bpPRD p_prd[8];

/*added by Jeff for ECC, 2001/11/19 */
	U8	blECCInitCompleted;
	U8	DoingECCInit;
/*END*/

 #ifdef _INVERT_
 	PVOID		pSource;
 	U8			Inverting;
 	U8			InvertType;
 	U8			FreeSEQNum;
 #endif
  
} PACKED SEQ, PACKED buf_space *bpSEQ, PACKED str_space *spSEQ;


#define SMALL_SCRATCH_BUFFER_SIZE      0x880 //0xa00//0xc40+sizeof(SEQ)*0xa*32 //0x1000 /* Modified by David Wei*/
#define XOR_PRD_BUFFER_SIZE            0x1200 //   0x2000

typedef struct sbuf {
  spQ  p_next;                                  /* link pointer */
  ByTE bReserved[8];
  union {
    bpQ  p_link;
    WORD w;
  } link;
  ByTE buf[SMALL_SCRATCH_BUFFER_SIZE];
} PACKED SBUF, PACKED buf_space *bpSBUF;


typedef struct sg {
  WORD wStartAddr;                              /* physical addr */
  WORD wCount;                                  /* size of element in bytes */
} PACKED SG, PACKED app_space *apSG, PACKED str_space *spSG;


/*
 * this is the main command packet which describes an OS request for disk service.
 * all pointer values are virtual address format except entries in SG table
 */

#define blOS_REQUEST                    0x00    /* bType */
						/* has bSTATUS_SEQ as last seq */
#define blENGINE_ACCESS                 bit(0)  /* has bSIGNAL_SEQ as last seq */
#define blCONVERTED_SG                  bit(1)  /* sg entry converted from virtual buf */
#define blSEGMENTED_REQUEST             bit(2)  /* REQ is split into multiple parts */
#define blNON_SECTOR_BOUNDARY           bit(3)  /* sg not on sector boundary */
#define blCONV_EXP_REQ                  bit(4)

#define blCONV_PBM_REQ                  bit(5)
/*#define blPBM_INTERNAL_REQ				bit(6)*/





#define bSIMPLE_TAG                     SRB_SIMPLE_TAG_REQUEST /* bTagAction */
#define bHEAD_OF_QUEUE                  SRB_HEAD_OF_QUEUE_TAG_REQUEST
#define bORDERED_TAG                    SRB_ORDERED_QUEUE_TAG_REQUEST

typedef struct sect {
    U32  wLBA;
    U16  hCount;
    bpPRD p_CurrentPRD;                            /* 03/23/98    Q.W */
    bpPRD pBlockFirstPRD;
    ByTE  NonOS;
    ByTE  bReserved;
    PRD   prd;                                    /* must be last */
} PACKED SECT, PACKED buf_space *bpSECT;

/*
typedef struct sect {
    WORD  wLBA;
    HALF  hCount;
    HALF  hReserved;
    PRD   prd; 
} SECT, buf_space *bpSECT;
*/

#if 0
typedef struct req {
  spQ    p_next;                                /* -> next blocked req */
  ByTE   bType;                                 /* RAID Engine only */
  ByTE   bAdapterStatus;                        /* init to asINCOMPLETE_REQ */
  ByTE   bTargetStatus;                         /* init to ssINCOMPLETE_CMD */
  ByTE   bID;                                   /* array id */
  ByTE   bLUN;                                  /* must be 0 */
  ByTE   bTag;                                  /* RAID Engine only */
  ByTE   bSG_Count;                             /* 0 = no sg, 1 and up = sg */
  ByTE   bSenseSize;                            /* >= 14 */
  WORD   wSize;                                 /* total size of xfer */
  spBUF  p_buf;                                 /* virtual -> a buffer or */
  spSG   p_sg;                                  /* virtual -> sg entries */
  spBUF  p_sns;                                 /* virutal -> sense data buffer */
  ByTE   bCDB[12];                              /* actual scsi cmd */
  void  *p_DeviceExtension;                     /* for NT */
  struct map buf_space *p_map;                  /* RAID Engine Only */
  bpSEQ  p_sq;                                  /* RAID Engine Only */
  bpQ    p_head[RESOURCE_FREE_POOL_COUNT];      /* RAID Engine Only */
  WORD   wLBA;                                  /* RAID Engine Only */
  HALF   hCount;                                /* RAID Engine Only */
  ByTE   bTagAction;                            /* for NT */
  ByTE   bReserved;
  void  *Srb;
} PACKED REQ, PACKED buf_space *bpREQ, PACKED str_space *spREQ;
#endif


/*
 * describes a disk access to channel driver.  all pointers are virtual address
 * except wPRD.  read/write commands access up to 64K. 
 */

#define DISK_TASK_TYPE                  0       /* bCh */

#define TASK_FLAG_MASK                  0xF0    /* bID, IOA only */
#define INT_WHEN_DONE                   bit(5)  /* IOA only */
#define LAST_TASK_OF_SAME_CH            bit(4)  /* IOA only */
#define SLAVE_DEVICE                    bit(0)

#define VIRTUAL_OPCODE                  bit(0)  /* blCmdFlag */
#define LBA2CHS                         bit(1)
#define DATA_IN                         bit(2)
#define DATA_XFER                       bit(3)
#define DMA_XFER                        bit(4)
#define ENGINE_DTASK                    bit(5)
#define NON_SECTOR_BOUNDARY             bit(6)

#define blRESET_ONCE                    bit(0)  /* blFlag */
#define blCONSOLIDATED_TASK             bit(1)
#define blENABLE_SMART					bit(2)
#define blSTARTBUSMASTER				bit(3)
typedef struct consolidate_task {
  spQ     p_next;                                 /* link pointer */
  ByTE    bReserved[3];
  ByTE    bTaskCount;
  BoOL    blOldFlag;
  ByTE    bOldCount;
  HALF    hOldSize;
  bpQ     p_link;
  bpPRD   p_prd;
  struct  dtask buf_space *p_tk[MAX_CONCURRENT_THREAD];
  PRD     prd;
} PACKED CTASK,  PACKED buf_space *bpCTASK;


typedef struct dtask {
  spQ    p_next;                                /* link pointer */
  ByTE   bCh;                                   /* channel number of disk */
  ByTE   bID;                                   /* 0 = master, 1 = slave */
  ByTE   bStatus;                               /* Bus Master status */
  ByTE   bMagic;                                /* signature byte to id struct */
  spSEQ  p_sq;                                  /* RAID Engine only */
  bpQ    p_link;                                /* duplicated ptr for release */
  WORD   wPRD;                                  /* physical -> PRD table */
  bpPRD  p_PRD;                                 /* virtual -> PRD table */
  /* Added by crimson. 5/26 */
  WORD   wPRD_Channel;                           /* physical -> PRD table for Channel */
  bpPRD  p_PRD_Channel;                          /* virtual -> PRD table for Channel */
  union {
    ByTE b1Fx[7];                               /* 1F1 to 1F7 */
    struct {
      #ifdef BIOS
      ByTE bFeature;                            /* ATA feature Register */
      ByTE bCount;                              /* # of sctr to process upto 128 */
      WORD wLBA;                                /* starting sector location */
      ByTE bCmd;                                /* ATA Command Register */
      ByTE bStatus;                             /* ATA Status Register */
      #else
      WORD wLBA;                                /* starting sector location */
      ByTE bFeature;                            /* ATA feature Register */
      ByTE bCount;                              /* # of sctr to process upto 128 */
      ByTE bCmd;                                /* ATA Command Register */
      ByTE bStatus;                             /* ATA Status Register */
      #endif
    } PACKED cmd;
  } ata_reg;
  HALF  hSize;                                  /* total xfer in half (2-byte) */
  BoOL  blCmdFlag;
  ByTE  bTimeout;                               /* in number of 128 msec */
  bpCTASK p_ct;
  ByTE  bRetry;
  BoOL  blFlag;
  BoOL  NonOS;
  ByTE  bReserved;
#ifdef __FreeBSD__
  void *buf;
#endif /* __FreeBSD__ */
} PACKED DTASK, PACKED buf_space *bpDTASK;


/*
 * describes a xor job.  all pointers are virtual, including pointers in PRD.
 */

#define XOR_TASK_TYPE                   bit(6)  /* bCh */

#define INT_WHEN_DONE                   bit(5)  /* bCount, IOA only */
#define LAST_TASK_OF_SAME_CH            bit(4)  /* IOA only */
#define XOR_COUNT_MASK                  0x07

typedef struct xtask {
  spQ    p_next;                                /* link pointer */
  ByTE   bCh;
  ByTE   bCount;
  ByTE   bStatus;
  ByTE   bMagic;                                /* signature byte to id struct */
  spSEQ  p_sq;                                  /* RAID Engine only */
  bpQ    p_link;                                /* duplicated ptr for release */
  bpPRD  p_PRD[MAX_STREAM_PER_ARRAY];           /* array of virtual -> PRD table */
} PACKED XTASK, PACKED buf_space *bpXTASK;

typedef struct readsg{                         /* used for read first block for XOR*/
    SG readSG;
    U8 RSG[READSG_BUFFER_SIZE];
} PACKED READSG, PACKED app_space *apREADSG, PACKED str_space *spREADSG;  

typedef struct readbuf{
    spQ  p_next;                                  /* link pointer */
    U8 bReserved[8];
    union {
        bpQ  p_link;
        U32 w;
    } link;
    U8 buf[READ_XOR_BUFFER_SIZE];
} PACKED READBUF, PACKED app_space *apREADBUF, PACKED str_space *spREADBUF;

/* 
    Added by crimson. 05/27. 
    Separate data block and control block for nonaccessable memory        
*/
typedef struct readbuf_control_block{
    spQ  p_next;                                  /* link pointer */
    U8 bReserved[8];
    union {
        bpQ  p_link;
        U32 w;
    } link;
    U32 buf;                                      /* Location in physical memory */  
} PACKED READBUF_CONTROL_BLOCK, PACKED app_space *apREADBUF_CONTROL_BLOCK, PACKED str_space *spREADBUF_CONTROL_BLOCK;


/* 
    Added by crimson. 05/27. 
    Separate data block and control block for nonaccessable memory        
*/
typedef struct readbuf_data_block{
    U8 buf[READ_XOR_BUFFER_SIZE];
} PACKED READBUF_DATA_BLOCK, PACKED app_space *apREADBUF_DATA_BLOCK, PACKED str_space *spREADBUF_DATA_BLOCK;


typedef struct paritybuf{
    spQ  p_next;                                  /* link pointer */
    U32  wlocation;                               /* pointer to physical address */
    U8   reserved[4];
    union {
        bpQ  p_link;
        U32 w;
    } link;
    U8 buf[PARITY_BUFFER_SIZE];
} PACKED PARITYBUF, PACKED app_space *apPARITYBUF, PACKED str_space *spPARITYBUF;

/* 
    Added by crimson. 05/27. 
    Separate data block and control block for nonaccessable memory        
*/
typedef struct paritybuf_control_block{
    spQ  p_next;                                  /* link pointer */
    U8 bReserved[8];
    union {
        bpQ  p_link;
        U32 w;
    } link;
    U32 buf;                                /* Location in physical memory */  
} PACKED PARITYBUF_CONTROL_BLOCK, PACKED app_space *apPARITYBUF_CONTROL_BLOCK, PACKED str_space *spPARITYBUF_CONTROL_BLOCK;


/* 
    Added by crimson. 05/27. 
    Separate data block and control block for nonaccessable memory        
*/
typedef struct paritybuf_data_block{
     U8 buf[PARITY_BUFFER_SIZE];
} PACKED PARITYBUF_DATA_BLOCK, PACKED app_space *apPARITYBUF_DATA_BLOCK, PACKED str_space *spPARITYBUF_DATA_BLOCK;


typedef struct paritysg{
    SG paritySG;
    U8 PSG[PARITYSG_BUFFER_SIZE];
} PACKED PARITYSG, PACKED app_space *apPARITYSG, PACKED str_space *spPARITYSG;  

typedef struct recoverbuf{
    spQ  p_next;                                  /* link pointer */
    U8 bReserved[8];
    union {
        bpQ  p_link;
        U32 w;
    } link;
    U8 buf[RECOVER_BUFFER_SIZE];
} PACKED RECOVERBUF, PACKED app_space *apRECOVERBUF, PACKED str_space *spRECOVERBUF;

/* 
    Added by crimson. 05/27. 
    Separate data block and control block for nonaccessable memory        
*/
typedef struct recoverbuf_control_block{
    spQ  p_next;                                  /* link pointer */
    U8 bReserved[8];
    union {
        bpQ  p_link;
        U32 w;
    } link;
    U32 buf;                                /* Location in physical memory */  
} PACKED RECOVERBUF_CONTROL_BLOCK, PACKED app_space *apRECOVERBUF_CONTROL_BLOCK, PACKED str_space *spRECOVERBUF_CONTROL_BLOCK;

/* 
    Added by crimson. 05/27. 
    Separate data block and control block for nonaccessable memory        
*/
typedef struct recoverbuf_data_block{
    U8 buf[RECOVER_BUFFER_SIZE];
} PACKED RECOVERBUF_DATA_BLOCK, PACKED app_space *apRECOVERBUF_DATA_BLOCK, PACKED str_space *spRECOVERBUF_DATA_BLOCK;


typedef struct recoversg{
    SG recoverSG;
    U8 PSG[RECOVERSG_BUFFER_SIZE];
} PACKED RECOVERSG, PACKED app_space *apRECOVERSG, PACKED str_space *spRECOVERSG;  


/*
typedef struct rebuildbuf{
  U8 buf[REBUILD_BUFFER_SIZE];
} PACKED REBUILDBUF, PACKED app_space *apREBUILDBUF, PACKED str_space *spREBUILDBUF;
)*/


/* 
    Added by crimson. 08/09, 2001. 
    Separate data block and control block for nonaccessable memory        
*/
typedef struct rebuild_control_block{
    spQ  p_next;                                  /* link pointer */
    U8 bReserved[8];
    union {
        bpQ  p_link;
        U32 w;
    } link;
    U32 buf;                                /* Location in physical memory */  
} PACKED REBUILDBUF_CONTROL_BLOCK, PACKED app_space *apREBUILDBUF_CONTROL_BLOCK, PACKED str_space *spREBUILDBUF_CONTROL_BLOCK;

/* 
    Added by crimson. 05/27. 
    Separate data block and control block for nonaccessable memory        
*/
typedef struct rebuild_data_block{
    U8 buf[REBUILD_BUFFER_SIZE];
} PACKED REBUILDBUF_DATA_BLOCK, PACKED app_space *apREBUILDBUF_DATA_BLOCK, PACKED str_space *spREBUILDBUF_DATA_BLOCK;



typedef struct rebuildsg{
    SG rebuildSG;
    U8 PSG[REBUILDSG_BUFFER_SIZE];
} PACKED REBUILDSG, PACKED app_space *apREBUILDSG, PACKED str_space *spREBUILDSG;  

typedef struct xormap{
    bpSECT p_sect[MAX_STREAM_PER_ARRAY1];
    SECT   sect;
}PACKED XORMAP, PACKED app_space *apXORMAP,PACKED buf_space *bpXORMAP;  

#define XORMAPLength    ((sizeof(SECT)+16*sizeof(PRD)+14)*(MAX_STREAM_PER_ARRAY1+1) )

typedef struct xormap1{
    bpSECT p_sect[MAX_STREAM_PER_ARRAY1];
    RECOVERSG recoversg;
    SECT   sect;
    /*U8 buf[640];*/
	/* Modified by crimson. 09/03, 2001. from david ...    	*/
    U8 buf[XORMAPLength];
}PACKED XORMAP1, PACKED app_space *apXORMAP1,PACKED buf_space *bpXORMAP1;  


typedef struct readmap{
    struct xormap buf_space *p_xormap;
    U32   rOffset;
#ifndef CACHE
    U32   wOffset;
    U32   wOffsetSG;
    apSG   p_current;                             /* read  current  sg pointer*/
    apSG   p_pcurrent;                            /* parity current sg pointer*/
    U16  hParityCount;
    apBUF  p_buffer;
#endif
    struct req *p_rq;
    apSG   p_sg;
    U8   bLogicalWidth;
    U8   bXorline;
    U16   hStripeBlockSize;
    BoOL   bReadPartOfBlock;
    BoOL   bReadTailPartOfBlock; 
    U16   hSizeOfPartBlock;
    U16   hTailSizeOfPartBlock;
    U8    xornum[32];  // David modifed it by Quincy direction
    U8   bRAID5Status;
    U8   bRMWStatus;
    U8   bAddedBlockSize;
    U8   bFirstLastSize;
    U8   bSecondLastSize;
    U8   bMinXorNum;
    U8   bMaxXorNum;
    U8   bRecoverStream;
    U8   bRebuildCount;
    U8   bPStream;
    BoOL  blFirstNotDone;
    U8   bConv_AryID;                /*  0: source array, 1: target array */
    U16  hOffset;
    U8   bRecoverCount1;
    U8   bRecoverCount2;
    U8   bBadSectorStream;
    U8   bWriteReqReadbackFailStream;
    U32  wLBAStart;       
    U32  wLBAEnd;
#ifdef CACHE
    CONTROL_BLOCK *pCB[40];
#endif
    PARITYSG  paritysg;
    READSG readsg;
    REBUILDSG rebuildsg;
    XORMAP1 xormap1;
    XORMAP1 xormap2;
    XORMAP xormap;
}PACKED READMAP, PACKED buf_space *bpREADMAP;


/*#define XOR_PRD_BUFFER_SIZE  (sizeof(READMAP)+((sizeof(SECT)+32*sizeof(PRD)+14)*(MAX_STREAM_PER_ARRAY1+1) )) Modified by David Wei 2001 08027*/


typedef struct xorbuf {
    spQ  p_next;                                  /* link pointer */
    U8 bReserved[8];
    union {
        bpQ  p_link;
        U32 w;
    } link;
    U8 buf[XOR_PRD_BUFFER_SIZE];
} PACKED XORBUF, PACKED buf_space *bpXORBUF;

typedef struct chprdtbl {
    spQ  p_next;                                  /* link pointer */
    U8 bReserved[8];
    union {
        bpQ  p_link;
        U32 w;
    } link;
    U8 buf[PRD_TBL_SIZE];
} PACKED CHPRDTBL, PACKED app_space *apCHPRDTBL, PACKED str_space *spCHPRDTBL;

typedef struct map {
    bpSECT p_sect[MAX_STREAM_PER_ARRAY];
    apSG   p_current;
    WORD   wOffset;
    WORD   wSkipSize;
    ByTE   bNumber;                               /* total number of stream involved */
    ByTE   bFirstStream;                          /* stream number of head */
    HALF   hReserved;
    SECT   sect;                                  /* must be last */ 
} PACKED MAP, PACKED buf_space *bpMAP;




typedef struct req {
    spQ    p_next;                                /* -> next blocked req */
    	/** added by Xinhai for new error recovery scheme */
	WORD wReqFlag;	                            /* Xinhai: see above blREQF_xxx */
	ByTE bErrorStatus;                          /* filled in ChannelReturn on disk error */
	ByTE bErrorValue;                           /* filled in ChannelReturn on disk error */
	ByTE bErrorCmd;                             /* error task command value, see IDE_COMMAND_xxx */
	ByTE bReserved0;
	WORD wErrorLBA;                             /* filled in ChannelReturn on disk error */
	WORD wErrorBlks;
    
    U8   bType;                                 /* RAID Engine only */
    U8   bAdapterStatus;                        /* init to asINCOMPLETE_REQ */
    U8   bTargetStatus;                         /* init to ssINCOMPLETE_CMD */
    U8   bID;                                   /* array id */
    U8   bLUN;                                  /* must be 0 */
    U8   bTag;                                  /* RAID Engine only */
    U8   bSG_Count;                             /* 0 = no sg, 1 and up = sg */
    U8   bSenseSize;                            /* >= 14 */
    U32   wSize;                                 /* total size of xfer */
#ifdef iRAID
    U16  scsiThreadIdx ;
#endif
    spBUF  p_buf;                                 /* virtual -> a buffer or */
    spSG   p_sg;                                  /* virtual -> sg entries */
#ifndef I2O
    spBUF  p_sns;                                 /* virutal -> sense data buffer */
#else
    U8   p_sns[14];
#endif
    U8   bCDB[12];                                                                  /* actual scsi cmd */
#ifndef I2O 
    void  *p_DeviceExtension;                     /* for NT */
#endif
    struct req *p_Lockrq;
    bpMAP  p_map;                  /* RAID Engine Only */
    bpREADMAP p_readmap;          /* RAID Engine Only */
    bpSEQ  p_sq;                                  /* RAID Engine Only 04/01/98 Q.W   */  
    bpQ    p_head[RESOURCE_FREE_POOL_COUNT];      /* RAID Engine Only */
    U32   wLBA;                                  /* RAID Engine Only */
    U16   hCount;                                /* RAID Engine Only */
#ifndef I2O
    U8   bTagAction;                            /* for NT */
#endif
    BoOL blConflict;
    BoOL blBadSector;
    void  *Srb;
    struct  req buf_space *p_rq;
    struct  event buf_space *p_evPBM_Req_Timedout;
#ifdef WINNT
#if defined (I2O) || defined (CACHE_EXIST)
    PARRAY_DEVICE                 pDevice;
#endif /*I20 || CACHE_EXIST*/
#endif /*NT*/

#ifdef  I2O
    spSG   p_SG;
    PI2O_SG_ELEMENT               pChain;
    PI2O_MESSAGE_FRAME    pMsgFrame;      
#endif
#ifdef CACHE_EXIST

#ifdef WINNT
    U32           *pSystemIrp;
#endif /*NT*/
    struct req    *pHR;                   /* For use by target request struct */
    struct req    *pNextLockReq;               /* -> next blocked req */
    struct _AVL_NODE *pCB;
    DMA_TYPE      CallBackFunction;
    U8            ReadAheadFlag;
    U8            SplitReqFlag;
    BIT_STATE     NextStateBusyBit;
    BIT_STATE     NextStateValidBit;
    BIT_STATE     NextStateDirtyBit;
    U16           ReadClaimCount ;
#ifdef I2O
    PI2O_SG_ELEMENT               pDMA_SG;
#endif /*I2O*/
    spSG                  pTR_SG;
    U8                   TR_bSG_Count;   /* babak: was U8 */
    U32                   TR_wSize;
    U32                   Req_LBA;
    U16                   Req_Count;
    U16                   OrigCount;
    U16                   ReadAheadCount;
    struct req    *pNextFree;
#ifdef DEBUG_MEM
    MEM_TAG_TYPE   MemTag ;
#endif
#endif

#ifdef _PBM_
    U32     flag;
    void    *ptr;			/* pointer to buf_req_t */

    void    *mem_p;         /* debug only */
    void    *sg;            /* debug only */
    U32     blkno;          /* debug only */
    U32     size;           /* debug only */
    U32     buf_flag;       /* debug only */

#ifdef NT_DEBUG
	unsigned char   debugFlag;
#endif

#endif /* _PBM_ */
#ifdef __FreeBSD__

    void *pccb;
#endif /* __FreeBSD__ */
#ifdef _INVERT_
	apSG  psg;
#endif

} PACKED REQ, PACKED buf_space *bpREQ, PACKED str_space *spREQ;


/*
 * stripe parsing stuff
 */
/*
typedef struct sect {
  WORD  wLBA;
  HALF  hCount;
  HALF  hReserved;
  PRD   prd;                                     
} SECT, buf_space *bpSECT;

*/

/*
 * hardwork stuff
 */

#define blVALID_EVENT                   bit(0)  /* blFlag */
#define blABORT_EVENT                   bit(1)
#define blPAUSE_EVENT					bit(2)
#define blACCESS_ONESECTOR				bit(3)
#define blECC_BREAK						bit(4)
#define blOSBUSY_EVENT                  bit(5)
#define blCONV_WRITE_ERR				bit(6)
#define blCONV_READ_ERR                 bit(7)

typedef struct event {
  spQ  p_next;
  BoOL blFlag;
  ByTE bFunction;
  ByTE bReserved;
  ByTE bMagic;
  HALF hETA;
  HALF hInterval;
  bpDTASK p_tk;
  WORD wStartLBA;
  WORD wEndLBA;
  union {
    struct {
      bpREQ   p_rq;
      ByTE    bAry;
      ByTE    bStrm;
      ByTE    bBlk;
    } PACKED recon;
    struct {
      bpREQ   p_rq;
      ByTE    bAry;
	  ByTE    bStrm;
	  ByTE    bBlk;
	} PACKED chkinte;
    struct {
      bpREQ   p_rq;
      ByTE    bAry;
	  ByTE    bStrm;
	  ByTE    bBlk;
    } PACKED pbm_req;
    struct {
      bpREQ   p_rq;
      ByTE    bCh;
      ByTE    bID;
      ByTE    bBlk;
      ByTE    bOffset;
      ByTE    bCount;
      ByTE    bOriginalOffset;
      HALF    hSize;
      apBUF   p_src;
    } PACKED config;
    struct {
      bpREQ   p_rq;
      ByTE    b0;
      ByTE    b1;
      ByTE    b2;
      ByTE    b3;
    } PACKED peg;
  } param;
} PACKED EVENT, PACKED buf_space *bpEVENT;


/*
 * locks sections
 */

#define blACCESS_LOCK                   bit(0)  /* blFlag */

typedef struct seg {
  spQ  p_next;
  WORD wStart;
  WORD wEnd;
  bpQ  p_link;
  BoOL blFlag;
  BoOL bAry;
  bpREQ p_rq;
} PACKED SEG, PACKED buf_space *bpSEG;

typedef struct logevent{
	WORD id;
	WORD TimeStamp;
	WORD QueueNum;
	WORD NumOfParams;
	WORD WarningLevel;
	union {
		struct {
			WORD wDiskID;                            /* ATA feature Register */
		}PACKED disk_offline;
		struct {
			WORD wDiskID;                            /* # of sctr to process upto 128 */
		}PACKED disk_setdown;
		struct {
			WORD wArrayID;                                /* starting sector location */
		}PACKED array_setdown;
		struct {
			WORD wArrayID;                                /* starting sector location */
		}PACKED array_degrade;
		struct {
			WORD wDiskID;                                /* starting sector location */
		}PACKED plug_in;
		struct {
			WORD wLBA;
			WORD wDiskID;
		}PACKED disk_retry;	
		struct {
			WORD wLBA;
			WORD wDiskID;
		}PACKED reserved_sector_error;	
		struct {
			WORD wLBA;
			WORD wDiskID;
			WORD wArrayID;
			WORD wStatus;
		}PACKED ecc_error;	
        struct {
            WORD wArrayID;                            /* ATA feature Register */
        }PACKED array_create;
        struct {
            WORD wArrayID;                            /* ATA feature Register */
        }PACKED array_delete;
        struct {
			WORD wArrayID;                            /* ATA feature Register */
			WORD wDiskID;
		}PACKED rebuild_start;
		struct {
			WORD wArrayID;
		}PACKED sync_start;
		struct {
			WORD wArrayID;                            /* ATA feature Register */
			WORD wDiskID;                              /* # of sctr to process upto 128 */
			WORD percent;                                /* ATA Command Register */
		}PACKED rebuild_complete;
		struct {
			WORD wArrayID;                            /* ATA feature Register */
			WORD percent;                                /* ATA Command Register */
		}PACKED sync_complete;
		struct {
			WORD wArrayID;                            /* ATA feature Register */
			WORD wDiskID;                              /* # of sctr to process upto 128 */
			WORD percent;                                /* ATA Command Register */
		}PACKED rebuild_abort;
		struct {
			WORD wArrayID;                            /* ATA feature Register */
			WORD percent;                                /* ATA Command Register */
		}PACKED sync_abort;
		struct {
			WORD wLBA;
			WORD wLBAEnd;
			WORD wArrayID;
		}PACKED sync_comp_error;
		struct {
			WORD wArrayID;                            /* ATA feature Register */
			WORD wDiskID;                              /* # of sctr to process upto 128 */	
			WORD wLBA;
			WORD wChannel;                                /* ATA Command Register */
			WORD wCmd;
			WORD wStatus;                             /* ATA Status Register */
		}PACKED disk_error;
		struct {
			WORD wDiskID;
		}PACKED smart_error;
		struct {
			WORD wChannel;
			WORD wID;
		}PACKED swap_box_error;
		struct {
			WORD wChannel;
			WORD wID;
			WORD value;   /* param value fan=rpm, 5v/1v=millivolts, temp=2*degC*/
		}PACKED swap_box_monitor_error;
	} ParamList;
}PACKED LOGEVENT, PACKED buf_space *bpLOGEVENT; 

typedef struct erlog{
	spQ p_next;
	WORD blFlag;
	LOGEVENT logEvent;
}PACKED ERLOG,PACKED buf_space *bpERLOG;

//Label_bp begin
//kernel modified for SuperSwap 1100/4100, define Swap Box Status;
#define INITCHECK     0
#define NODISKBOX		0x80	// bit(7)=1 nobox exist
#define HAS_VOLTAGE		0x04	// bit(2)=0 with valtage sensor
#define HAS_FAN_TEMP	0x08	// bit(3)=0 with fan and temperature
#define HAS_KEY			0x40	// bit(6)=0 with key

//kernel modified for SuperSwap 1100/4100, define Swap Box Type
#define FastSwap66		0x01	// 00000001 b
#define SuperSwap		0x11	// 00010001 b
#define OEMGigabyte		0x10	// 00010000 b
#define OEMuniversal	0x4C	// 0x00xx00 b
#define SuperSwap1100	0x00	// 10000000b
#define SuperSwap4100	0x09	// 10000009b
#define NOBOX			0xFF	// 11111111 b
#define UNKNOWNBOX		0xFF	// 11111111 b
//Label_bp end


//modified by isaac for dualcard
typedef struct enclosure{
	ByTE ID[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
    ByTE IDStatus[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
	ByTE LEDStatus[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
	ByTE LEDColor[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
	WORD TEMPStatus[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
	WORD TEMPStatus2[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
	WORD TEMPStatus3[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
	WORD FANStatus[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
	WORD FANStatus2[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
	WORD FANStatus3[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
	WORD VOL5V[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
	WORD VOL12V[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
	ByTE NeedCheck[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
    ByTE HasWinBond[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
    ByTE PowerState[MAX_CHANNEL_SUPPORTED][MAX_ID_PER_CHANNEL];
    ByTE PWMTable[MAX_CHANNEL_SUPPORTED];
    //Label_bp begin
    #ifdef HPQ_BACKPLANE    
    ByTE GlobalLED[MAX_ADATPER_SUPPORTED];
	BoOL isIdentifyDriveEnabled[MAX_ADATPER_SUPPORTED];
	//Label_bp end
#endif
}PACKED ENCLOSURE, PACKED *pENCLOSURE;


/***************************************************************************
 *
 * miscellaneous
 *
 ***************************************************************************/

typedef struct mode_page {
  ByTE bPage;                                   /* page code */
  ByTE bSize;                                   /* size of page */
  void (* GetPage)   (ByTE bType, apBUF p_buf, bpREQ p_rq);
  ByTE (* bPutPage)  (apBUF p_buf);
} PACKED MODE_PAGE, PACKED app_space *apMODE_PAGE;


/*#ifdef NT_MSC
/#pragma pack (4)
/#endif*/
#ifdef NT
#pragma pack () /* 12/17/97 S.Zhu*/
#endif


