/*********************************************************************
* os.c
* 10/30/00 Linus Chen
*       Moved all the functions exported by the os driver to this 
*	file, and declare array to save SCSI commands instead of 
*	allocating memory dynamically. 
* 11/08/00 Linus Chen
*	Fix SwapBox power control 
* 11/24/00 Frank Tiernan
*	Add reset_controller() to fix bonnie problem called by
*	fasttrak_reset().And Adding 'pending' member to struct REQ_head.
* 12/12/00 Linus Chen
*	Remove ide.h and scsi.h to compile as independent object 
* 02/06/01 Linus Chen	v1.20 build 3
*	Add verification of Promise controller on "SCSI INQUIRY" 
*	to prevent from the partial source released to be used on
*	other IDE controller.
*********************************************************************/
 
#include <linux/blkdev.h>
#include <asm/string.h>
#include "global.h"
#include "wrapper.h"
//Label5 begin
#include "engine/oem.h"
//Label5 end
#include "engine/code.h"
#include "engine/constant.h"
#include "engine/struct.h" 	//for bpDTASK structure
#include "fasttrak.h"


#include "engine/export.h"
#include "engine/ata-3.h"
#include "engine/global.h"
#include "engine/scsi-3.h"
#include "cam_con.h"
#include "cam_def.h"
#include "cam_macr.h"
#include "cam_var.h"
#include "ata.h"

#include "engine/macro.h"


#define	DID_OK		0
#define DID_ABORT	5

static unchar	INestcli=0 ; 
extern BoOL CheckEnclosureGlobalStatus(UCHAR bAdapterNumber, UCHAR bStatus);

ulong	SaveFlags;
struct timer_list	submit_timer;	// for submit_timer_handler
unchar	ReadCapacityCompleted ; // wait RAID_Return for READ_CAPACITY cmd
ulong	StartRebJiffies;	// start rebuild jiffies
unchar	DisMirrorSync   = 0 ;	// default = 0 : real time to rebuilding
unchar	SCSIRequestFlag = 0 ;
struct	REQ_head {
   REQ			req;
   SG			prdt[Max_SG_Table];
   void			*cmd;
   void			(*fn)(void *) ;
   struct REQ_head	*next;
   unsigned char        pending;
} ; 
typedef	struct	REQ_head REQ_Head_t ;
REQ_Head_t	Queue[Can_Queue+1], *PoolH, *PoolF ;	// Head/Foot
U16 	PTI_internal_cmd(REQ_Head_t *myqueue);
U16 	PTI_rw_cmd(REQ_Head_t *myqueue);
extern int buf_submit(REQ *req);

extern ByTE AlarmBuzzerEnabled;

//Label_bp begin
extern int g_Is_ShutDown;
extern U16 PTIAdapter;

extern void CheckSBoxEnclosureStaus(void);
extern BoOL ControlActivityLED(UCHAR bCh,UCHAR bID, UCHAR bOn);
void CheckStatusLEDChange(
	U8 bCh,
	U8 bID,	/* bID will be ignore */
	U8 bForceCheck
	);
extern ByTE GetFlashControl(ByTE status);
extern void ControlStatusLED(UCHAR bCh,UCHAR bID);
//Label_bp end

/*********************************************************************
 * Function:
 *   submit_timer_handler sub-routines :
 *	char PreReb(unchar array)
 *	char PreNotReb(unchar array)
 *
 * Description:
 *   	PreReb    - Array Rebuilding on last submit timer handler
 *	PreNotReb - Array Not Rebuilding on last submit timer handler
 *	   return - represent the Array is rebuilding or not
 *	        0 - Not Rebuilding
 *		1 - Rebuilding
*********************************************************************/
//int RebuildingInProgress=0;
//int ConversionInProgress=0;
extern void CamSwapLight(U8,U8,U8,U8);

//added by isaac for Synchronize
//extern BoOL SynchronizeComplete; 
//extern BoOL ConExpComplete;
extern U8 MaintenanceInProgress;
U8 PreReb(U8 array)
{
   static int	percent ;
   U8 	dev;
   // find previous rebuilding drive : dev
   
   for( dev=0 ; dev < PTI_Max_Dev ; dev++ ) {
	PTI_drive_t	*dp = &PTI_drive[dev] ;
	if( (dp->arr == array) && (dp->rebuilding) ) 
	   break;
   }		
   if( p_evMaintenanceEvent[array] ) {
	unsigned char arr     = PTI_drive[dev].arr ;	
	unsigned char strm    = PTI_drive[dev].strm ;
	unsigned char block   = aryDef[arr].strm[strm].bIndex; 
	unsigned long wSize   = blkDisk[block].wSize / 2048 ;
	unsigned long wWater  = blkDisk[block].wWatermark / 2048;
	percent = wWater * 100 / wSize ;
	return(1);
   }
  // printk("<1> aryDef[%d].blFlag=%x\n",array,aryDef[array].blFlag);
  // added by isaac for Maintenance
   if((dev == PTI_Max_Dev) && !(aryDef[array].blFlag&0x10) ){
	   
	   switch (MaintenanceInProgress & 0xff)
            {
	     case 0x02:
  	          log_event(4144,array,0,0,0,0,0);
	       	
		  //modified by isaac for alarm buzzer
                  //if(AlarmBuzzerEnabled) buzzer_off();
	          
		  MaintenanceInProgress &= ~SYNCHRONIZATION_IN_PROGRESS;
                  break;		  
	     case 0x04:
	          log_event(4142,array,0,0,0,0,0);
	       	
		  //modified by isaac for alarm buzzer
                  //if(AlarmBuzzerEnabled) buzzer_off();
	        
	          MaintenanceInProgress &= ~CONVEXP_IN_PROGRESS;
		  break;
	     case 0x08: 	  
                  log_event(4148,array,0,0,0,0,0);
	       	
		  //modified by isaac for alarm buzzer
              //    if(AlarmBuzzerEnabled) buzzer_off();
	          
	      	  MaintenanceInProgress &= ~RAID5_INITIALIZATION_IN_PROGRESS;
                  break;
	     case 0x22:
                  log_event(4146,array,0,0,0,0,0);
	          MaintenanceInProgress &= ~SYNCHRONIZATION_IN_PROGRESS;
                  break;
             case 0x44:
	          log_event(4145,array,0,0,0,0,0);
		  
		  //modified by isaac for alarm buzzer
                  //if(AlarmBuzzerEnabled) buzzer_on();
	          
		  MaintenanceInProgress &= ~CONVEXP_IN_PROGRESS;
                  break;
             case 0x88:
                  log_event(4149,array,0,0,0,0,0);		  
	          MaintenanceInProgress &= ~RAID5_INITIALIZATION_IN_PROGRESS;
                  break;
	     default:
	          PTI_printk("<1>Maintenance Event Flag Error! \n"); 
	   }	  
#if 0
   	   switch (aryDef[array].blFlag&0x60)
	{
	 case 0x60: 
	   // conversion complete
	     log_event(4142,array,0,0,0,0,0);
//	     ConversionInProgress = 0;
//	     ConExpComplete = FALSE;   
	     MaintenanceInProgress ~ = CONVEXP_IN_PROGRESS;
	     break;
	 case 0x40:       	
           //conversion stopped
	     log_event(4145,array,0,0,0,0,0);
//	     ConversionInProgress = 0;
//	     MaintenanceComplete = FALSE;
	     break;
	 case 0x00:
	    //synchronization complete
           if(SynchronizeComplete)
	   {
  	     log_event(4144,array,0,0,0,0,0);
//	     ConversionInProgress = 0;
	     SynchronizeComplete = FALSE;
	     MaintainanceComplete =TRUE;
	     break;
	   }
//	 default:
	    //synchronization stopped or Conversion Complete
	   else if(ConversionInProgress)
	   {
	     log_event(4142,array,0,0,0,0,0);
//	     ConversionInProgress = 0;
	     MaintainanceComplete = TRUE;
	   }
	   else
	   {
             log_event(4146,array,0,0,0,0,0);
//	     ConversionInProgress = 0;
	     MaintainanceComplete = FALSE;
	     break;
	   }
	 }	 
#endif  
         	 
   	// Define new field value of PTI_drive[dev] .arr & .strm
   	if( DefDriveStruct() ) {	
   	   PTI_printk("FastTrak: Define Drive Structure Error..\n");
   	   return(0);
   	}
//	ConversionInProgress=0;
   } 
          
   
   
  //modified by isaac for bug	
   else if(dev != PTI_Max_Dev){
 	
	// Rebuilding is stopped
   	PTI_drive[dev].rebuilding = 0;
   	if( aryDef[array].blFlag & 0x10 ) {
	//	*** Rebuild aborted ***
	//	The source disk maybe has been powered off
	//            or encountered a bad block
		// modified by isaac for dualcard           
		CamSwapLight(dev/8,(dev%8)/2,dev%2,3);         //set LED orange
	        log_event(4128, array, dev, 0,percent,0,0); 	
	       	
		//modified by isaac for alarm buzzer
                if(AlarmBuzzerEnabled) buzzer_on();
		
//		RebuildingInProgress=0;
		MaintenanceInProgress &= ~REBUILD_IN_PROGRESS;
	}
	else { 	// complete : 0x0F , fail : 0x1F
		// 	*** Rebuild successful ***
		// modified by isaac for dualcard	
		CamSwapLight(dev/8,(dev%8)/2,dev%2,2);         //set LED green 
        	log_event(4126, array, dev, 0,0,0,0); 
               
	       	//modified by isaac for alarm buzzer
                if(AlarmBuzzerEnabled) buzzer_off();

		//		RebuildingInProgress=0;
		MaintenanceInProgress &= ~REBUILD_IN_PROGRESS;
   	}
   }
/*  else {
        //conversion stopped
	  log_event(4145,array,0,0,0,0,0);
	  ConversionInProgress=0;
  }*/
   StartRebJiffies = 0 ;
   return(0);	
}
// -----------------------------------------------
char PreNotReb(unchar array)
{
   unchar dev, uc;
 
   if( p_evMaintenanceEvent[array] == NULL ) 
   	return(0);
   //	*** Start Rebuilding ***
   StartRebJiffies = PTI_get_jiffies() ;
   // Define new field value of PTI_drive[dev] .arr & .strm
   if( DefDriveStruct() ) {	
   	PTI_printk("FastTrak: Define Drive Structure Error..\n");
   	return(0);
   }
   // find rebuilding drive on aryDef[array]
   for( uc=0 ; uc < MAX_STREAM_PER_ARRAY ; uc++ ) {
   	dev = 0 ;
	if( aryDef[array].strm[uc].blFlag == 0x27 ) {
	   do {
	   	PTI_drive_t	*dp = &PTI_drive[dev] ;
		if((dp->present)&&(dp->arr == array)&&(dp->strm == uc)) 
		{
                   log_event(4103, array, dev,0,0,0,0);
		   
		   //modified by isaac for alarm buzzer
                   if(AlarmBuzzerEnabled) buzzer_off();
	   	   
		   dp->rebuilding = 1 ;
  //                 RebuildingInProgress=1;
		   
		   MaintenanceInProgress &= ~REBUILD_STOP;
		   MaintenanceInProgress |= REBUILD_IN_PROGRESS;
	   	   break;
		}
	   }
	    while( dev++ < PTI_Max_Dev ) ;
	   //modified by isaac for dualcard
     	   CamSwapLight(dev/8,(dev%8)/2,dev%2,3);         //set LED orange
	   return(1);
	}
   }	
   //modified by isaac for ConExp
   if((U32)p_evMaintenanceEvent[array] == 0xffffffff) 
   {
	//log_event(4141,array,0,0,0,0,0);	 
        //ConversionInProgress=1;
	   return(0);
   }   
   if(  p_evMaintenanceEvent[array]->bFunction == bCHECK_INTEGRITY)
   {
	log_event(4104,array,0,0,0,0,0);	 
        //modified by isaac for alarm buzzer
      //  if(AlarmBuzzerEnabled) buzzer_on();
	
        MaintenanceInProgress &= ~SYNCHRONIZATION_STOP;	
        MaintenanceInProgress |= SYNCHRONIZATION_IN_PROGRESS;	
	return(1);	
   }
 // if( aryDef[array].blFlag & 0x40) { 
   //modified by isaac according Edl's suggestion
   if( p_evMaintenanceEvent[array]->bFunction == bCONVEXP_RAID)
   {
        log_event(4141,array,0,0,0,0,0);	
         	
        //modified by isaac for alarm buzzer
        if(AlarmBuzzerEnabled) buzzer_off();
        
	MaintenanceInProgress &= ~CONVEXP_STOP;
        MaintenanceInProgress |= CONVEXP_IN_PROGRESS;
  //      ConversionInProgress=1;
   	return(1);	
   }
   //modified by isaac for Init RAID5 cleardisk event log
   if( p_evMaintenanceEvent[array]->bFunction == bCLEARDISK_RAID5)
   {
        log_event(4147,array,0,0,0,0,0);	
        
	//modified by isaac for alarm buzzer
        //if(AlarmBuzzerEnabled) buzzer_on();
        
	MaintenanceInProgress &= ~RAID5_INITIALIZATION_STOP;
        MaintenanceInProgress |= RAID5_INITIALIZATION_IN_PROGRESS;
  //      ConversionInProgress=1;
   	return(1);	
   }
  
   return(1);	
}		

extern BoOL CheckEnclosureIfDiskIdle(UCHAR bCh,UCHAR bID);
extern void SendSMARTCMD(void);
/*********************************************************************
 * Function:
 *   Timer Handler :
 *	void submit_timer_handler(ulong data)
 *	void do_submit_timer_handler(ulong data)
 *	void error_handler(ulong dev)
 *
 * Description:
 *   	submit_timer_handler : initialize in SubmitTimer(bpEVENT)
 *	error_handler        : initialize in SumitChannel(bpDTASK)
*********************************************************************/

void do_DiskCheck(void)
{
   int i,j,status;
   for (i=0;i<PTI_Max_Dev;i+=2)
   {
	if(PTI_drive[i].present==1){
	    if(DEVBUSY[i]<PTI_get_jiffies()-300){
		DEVBUSY[i]=PTI_get_jiffies();
		if(ataWaitingISR[i/2]==0){
			ataWaitingISR[i/2]=1;
			WriteIDEIndexUchar(gAdapterConfigInfo.ATABaseAddress[i/8],(i%8)/2,iNDEXIDEDeviceHead,(U8)(((i&0x03)<<04)|0xA0));
   			for(j=0;j<10;j++) {
   				PTI_udelay(100); 		      	
				ReadIDEIndexUchar(gAdapterConfigInfo.ATABaseAddress[i/8],(i%8)/2,iNDEXIDEAlternateStatus,(unsigned char *)&status);
				if( j == 9 ){
					PTI_drive[i].present=0;
					ReviseStructure(i/2);
				        log_event(4123, 0, i, 0, 0, 0, 0);
				}
				if( ((unsigned char)status == 0x7f)||((unsigned char)status == 0x00) ) continue;
				if((status & 0xD0)==0x50) 
					break;
   			}
			ataWaitingISR[i/2]=0;
			//Label_bp begin
			CheckStatusLEDChange(0xff,0,TRUE);
			//Label_bp end
		}
	    }
	}
   }	   
}

//Label_bp begin
void PlayWithLED(void)
{
	static U8 LEDCTLtimer=21;
	const U8 LEDCTLtimerLimit=3;
	int i=0;

	if( g_Is_ShutDown )
		return ;
	if (FastTrakAdapterType[i]  == RAINMAKER_621)
	{
	//added by Jeff for LED control
	LEDCTLtimer++;
	if (LEDCTLtimer > LEDCTLtimerLimit ) {
		U8  cnt;

//		if (CntForLED_in_timerservice==CntForLED_in_IOCTL){
			for (cnt=0;cnt<MAX_CHANNEL_SUPPORTED;cnt++){
				if (LEDCTLtimer==22)
					pEnclosure->LEDColor[cnt][0]=LED_GREEN;

				CheckEnclosureIfDiskIdle(cnt, 0);
			}
//		}
		//CntForLED_in_timerservice=	CntForLED_in_IOCTL;
		LEDCTLtimer=0;
	}
//end
		}
	else
	{
		CheckStatusLEDChange(0xff,0,FALSE);
#if !defined(_NEC)
		{
			CheckSBoxEnclosureStaus();
		}
#endif
	}
	{
		U8 ii, cnt, status[2] = {0,0}; // for dual card
			
		for (ii=0; ii< MAX_ARRAY_SUPPORTED; ii++)
			if( ARRAY_IN_REBUILDING(ii) )
				status[ii/4] |= SW_Rebuild;

		for (cnt=0; cnt<MAX_CHANNEL_SUPPORTED; cnt++)
		{
			if( DISK_ACCESS[cnt/4][cnt%4] == HDActiveON )
			{
				status[cnt/4] |= SW_DiskAccessed;
				ControlActivityLED(cnt, 0, HDActiveON);
			}
			else
				ControlActivityLED(cnt, 0, HDActiveOFF);
			
			DISK_ACCESS[cnt/4][cnt%4] = HDActiveOFF;
		}
		#if defined(_NEC)
		for (cnt=0; cnt<2; cnt++)
			CheckEnclosureGlobalStatus(cnt, status[cnt]);
		#endif
	}
}

#ifndef HPQ_BACKPLANE
U8 CollectInformationForStatusLED(
	U8 bCh,
	U8 bID,
	U8* flash
	)
{
	*flash = NOFLASH;

	if( g_Is_ShutDown )
	{
		return LED_OFF;
	}

	if( (blDiskFlag[bCh][bID]&blDISK_DOWN_BY_ERROR) )
	{
		/* Drive fails or offline, engine set down the disk because of error */
		return LED_RED;
	}
	if( !(blDiskFlag[bCh][bID]&blDISK_IS_PRESENT_HERE) )
	{
		/* Drive not installed */
//#if defined(_NEC)
      //Label4 begin
		return LED_ORANGE; // LED_OFF;	/* NEC asked for change */
	 //Label4 end	
//#else
//		return LED_RED;	/* 10398A cannot turn off LED, use RED instead */
//#endif
	}
#if !defined(_NEC)
	if( (SmartStatus[bCh][bID]==SMART_ERROR)	/* smart error */
//		|| () /* 5V error*/
//		|| () /* 12V error*/
//		|| () /* FAN error*/
//		|| () /* temp error*/
		)
	{
		return LED_ORANGE;
	}
#endif
	{
		U8 bStream, bBlk, bArray, blFlag;

		bBlk = bFindAssociatedBlock(0,bCh,bID);
		if( bBlk<MAX_BLOCKS )
		{
			bArray = bFindAssociatedArray(bBlk);
			if( bArray<ACTUAL_ARRAY_SUPPORTED )
			{
				if( !(aryDef[bArray].blFlag&blCONV_EXP_LUN) )
				{
					if( ARRAY_IN_REBUILDING(bArray) )
					{
						bpEVENT p_ev = p_evMaintenanceEvent[bArray];
						
						if( aryDef[bArray].strm[p_ev->param.recon.bStrm].bIndex
							== bBlk )
						{
							*flash = SLOWFLASH;
							return LED_ORANGE;
						}
					}
				}
				for( bStream=0; bStream<aryDef[bArray].bWidth; bStream++ )
				{
					dpSTRM p_sm = &aryDef[bArray].strm[bStream];

					if( p_sm->blFlag&blASSIGNED_BLK )
					{
						bpBLOCK p_blk = &blkDisk[p_sm->bIndex];

						if( p_blk->bCh==bCh && p_blk->bID==bID )
						{
							blFlag = p_sm->blFlag & ~(blASSIGNED_BLK|blREDIRECT_BLK|blDUPLICATE_BLK);
							if( blFlag!=blVALID_BLK+blWORKING_BLK )
							{
								//TRACE(("blFlag = %x\n",blFlag));
								return LED_RED;
							}
							break;
						}
					}
				}
			}
			blFlag = blkDisk[bBlk].blFlag & ~(blASSIGNED_BLK|blREDIRECT_BLK|blDUPLICATE_BLK);
			if( blFlag!=blVALID_BLK+blWORKING_BLK )
			{
			//	TRACE(("blFlag = %x\n",blFlag));
				return LED_RED;
			}
		}
	}

	/* everything is ok, return GREEN */
	return LED_GREEN;
}

void CheckStatusLEDChange(
	U8 bCh,
	U8 bID,	/* bID will be ignore */
	U8 bForceCheck
	)
{
	U8 startChannel, endChannel;

	if( bCh==0xff )
	{
		startChannel = 0;
		endChannel = MAX_CHANNEL_SUPPORTED - 1;
	}
	else
	{
		startChannel = bCh;
		endChannel = bCh;
	}

	for( bCh=startChannel; bCh<=endChannel; bCh++ )
	{

		if( bForceCheck )
		{
			U8 flash;
			U8 LedColor = CollectInformationForStatusLED(bCh,0,&flash);
			flash = GetFlashControl(flash);
			if( (LedColor==pEnclosure->LEDColor[bCh][0])
				&& ((pEnclosure->LEDStatus[bCh][0]&0xfc)!=flash) 
				&& !(pEnclosure->LEDStatus[bCh][0]&LED_CHANGE)
				)
			{
				/* nothing changed */
				break;
			}
			pEnclosure->LEDColor[bCh][0] = LedColor;
			pEnclosure->LEDStatus[bCh][0] &= 0xfc;
			pEnclosure->LEDStatus[bCh][0] |= (flash|LED_CHANGE);
		}
		if( (pEnclosure->LEDStatus[bCh][0]&LED_CHANGE) )
			{	
        		ControlStatusLED(bCh,bID);
			}
	}
}
#endif

extern PTI_drive_t 	PTI_drive[PTI_Max_Dev];
BoOL IsDriveThere(U8 bTargetID)
{
	return PTI_drive[bTargetID].present;
}
//Label_bp end

extern U16 PTIAdapter;
void do_submit_timer_handler(ulong data)
{
   unchar		array ;	
   bpEVENT		p_ev ;
   static unchar 	reb_arys[PTI_Max_Arr] = { 0,0,0,0,0,0,0,0 } ;
   static int ptiSendSmartCmd=0;
   DisableInt();
   // Put information to system log if Array status is 
   // different with last submit timer.
   DebEngFun(printk("enter submit timer handler, data=%x\n",data));
   
   //added by isaac for alarm buzzer
   if(!AlarmBuzzerEnabled) buzzer_off();
	   
   for( array=0 ; array < PTI_Max_Arr ; array++ ) 
	reb_arys[array] = (reb_arys[array])?PreReb(array):PreNotReb(array);
   if(ptiSendSmartCmd>=20){
        SendSMARTCMD();
	ptiSendSmartCmd=0;
        do_DiskCheck();
   }
   else
	ptiSendSmartCmd++;

//Label_bp begin
	PlayWithLED();
/*
{
	int i;
		 for(i=0;i<PTI_Max_Dev;i++){
			 //added by isaac for dualcard
			if( (PTIAdapter == 1)&&(i >= PTI_Max_Dev/2)) continue; 
			CheckEnclosureIfDiskIdle(i/2,i%2);
		}
}*/
//Label_bp end

   p_ev  = (bpEVENT) data ;
   hTicksTillEvent = 0 ; 
   TimerReturn(p_ev);
  

   EnableInt();
   return ;
} 
/***********************************************************
 * Function:
 *   void submit_timer_handler(ulong data)
***********************************************************/
extern int	CYC_ISR ;		// Linus Chen
int		CYC_TIMER = 0 ;
void submit_timer_handler(ulong data)
{
   ulong	smpflag;
   if( CYC_ISR | CYC_TIMER ) {
//   	if( CYC_ISR )
//	   printk("submit_timer_handler : CYC_ISR = 1 \n");
//   	if( CYC_TIMER )
//	   printk("submit_timer_handler : CYC_TIMER = 1 \n");
   	PTI_del_timer((struct PTI_timer_list* )(&submit_timer));
   	init_timer(&submit_timer);
   	submit_timer.expires = PTI_get_jiffies() + 50 ;	// delay 0.5 second
   	submit_timer.data = ( ulong ) data ;
   	submit_timer.function = submit_timer_handler ;
   	PTI_add_timer((struct PTI_timer_list* )(&submit_timer));
	return ;
   }	

   CYC_TIMER = 1 ;
//   spin_lock_irq(&PTI_adapter[0].pti_lock);
   spin_lock_irqsave(&io_request_lock,smpflag);
   do_submit_timer_handler(data);
   spin_unlock_irqrestore(&io_request_lock,smpflag);
//   spin_unlock_irq(&PTI_adapter[0].pti_lock);
   CYC_TIMER = 0 ;
   return;
}	
/***********************************************************
 * Function:
 *   void InitQueueBuffer(void)
***********************************************************/
void InitQueueBuffer (void) {
   unchar	queue;
   for( queue=0 ; queue <= Can_Queue ; queue++ ) {
	if( queue == Can_Queue )
	   Queue[queue].next = NULL ; 
	else
	   Queue[queue].next = &Queue[queue+1] ; 
   }
   PoolH = &Queue[0] ;
   PoolF = &Queue[Can_Queue] ;
   return;
}


U16 PTI_default_cmd(REQ_Head_t *myqueue)
{
   buf_submit(&(myqueue->req)) ;   
   return 0;
}


/***********************************************************
 * Function:
 *   do_queuecommand
 *
 * Description:
 *   cmd - SCSI command.
 *   fn  - function called on completion.
***********************************************************/
extern char EndRequest(void *cmd,void *fn,int result) ;
int pti_do_queuecommand(ScsiEngReq_t *CurScsiReq) 
{
   unchar		max_cdb_len = 12;
   ushort		us;
   REQ_Head_t		*myqueue;
   extern REQ		*shuttingREQ ;
   
   if( shuttingREQ ) {		// Linus Chen : for Synchronize Cache
	PTI_printk("FastTrak: PBM Shut Request !!\n");
   	EndRequest(CurScsiReq->cmd,CurScsiReq->fn,DID_ABORT << 16);
	return(0);
   }
   
   myqueue 	= PoolH ;
   PoolH	= myqueue->next;
   memset(myqueue,0,sizeof(REQ_Head_t));
   myqueue->req.flag=0;
   myqueue->cmd                 = CurScsiReq->cmd ;
   myqueue->fn                  = CurScsiReq->fn  ;
   myqueue->req.bAdapterStatus 	= 1 ;
   myqueue->req.bTargetStatus  	= 0 ;
   myqueue->req.bID     	= CurScsiReq->target ;
   myqueue->req.bLUN    	= 0 ;
   // a number of PRD Table entries
   myqueue->req.bSG_Count      	= CurScsiReq->use_sg ;
   myqueue->req.bSenseSize     	= 16 ;
   myqueue->req.p_buf          	= CurScsiReq->buffer ;
   myqueue->req.p_sns          	= CurScsiReq->sense ;

   for( us=0 ; us < max_cdb_len ; us++ ) {
   	myqueue->req.bCDB[us] = CurScsiReq->cmnd[us] ;
	if(myqueue->req.bCDB[0]==0x35);
   }
   if(CurScsiReq->use_sg) {	
	for( us=0 ; us < CurScsiReq->use_sg ; us++ ) {
	   myqueue->prdt[us].wStartAddr = CurScsiReq->WStartAddr[us] ;
	   myqueue->prdt[us].wCount     = CurScsiReq->WCount[us] ;
	}   
        myqueue->prdt[CurScsiReq->use_sg-1].wCount |= (1<<31) ;	// EOT=1
   	myqueue->req.p_sg = &myqueue->prdt[0] ;
   }
   SCSIRequestFlag ++ ;
   DBO(printk("<1>pti_do_queuecommand:receive scsi cmd 0x%x\n",myqueue->req.bCDB[0]));
   switch (myqueue->req.bCDB[0]) {
     	case scsiTEST_UNIT_READY :
	case scsiINQUIRY :
     	case scsiREAD_CAPACITY:
		return(PTI_internal_cmd(myqueue));
	case scsiREAD_6:
     	case scsiREAD_10:
     	case scsiWRITE_6:
     	case scsiWRITE_10:
  	   	return(PTI_rw_cmd(myqueue)); 		
     	default:
		PTI_printk("FastTrak: Unknown SCSI CMD(%#x)\n",myqueue->req.bCDB[0]);
	   	return(PTI_default_cmd(myqueue));
   }
   EndRequest(myqueue->cmd,myqueue->fn,DID_ABORT << 16);
   return 0;
}
// -----------------------------------------------

U16 PTI_internal_cmd(REQ_Head_t *myqueue)
{
    switch (myqueue->req.bCDB[0]) {
     	case scsiTEST_UNIT_READY :
   	   buf_submit(&(myqueue->req)) ;   
	   break;
	case scsiINQUIRY :
	   myqueue->req.wSize = myqueue->req.bCDB[4] ;
   	   buf_submit(&(myqueue->req)) ;   
	   break;
     	case scsiREAD_CAPACITY:
	   {
   	// when conversion, engine will not process this command. 
	// so it has been done by myself.
	   
	        unchar	ID   = (unchar  )myqueue->req.bID;
		unchar	*buf = (unchar *)myqueue->req.p_buf;
   		ulong	wSize  ;
		wSize = aryDef[ID].wSize - 1 ;
	   	buf[0] = (unchar) (wSize >> 24) ;
	   	buf[1] = (unchar) (wSize >> 16) ;
	   	buf[2] = (unchar) (wSize >>  8) ;
	   	buf[3] = (unchar)  wSize ;
	   	buf[4] = 0 ;
	   	buf[5] = 0 ;
	   	buf[6] = 0x02 ;
	   	buf[7] = 0 ;
	   }
   	   SCSIRequestFlag -- ;
   	   PoolF->next = myqueue ;
   	   PoolF = myqueue ;
   	   PoolF->next = NULL ;
   	   EndRequest(myqueue->cmd,myqueue->fn,(DID_OK << 16));
	   break ;
   }
   return 0;
}
extern ulong wV2P2(unchar *p_b);
// -----------------------------------------------
U16 PTI_rw_cmd(REQ_Head_t *myqueue)
{
   unchar cmnd4 = myqueue->req.bCDB[4];
   unchar cmnd7 = myqueue->req.bCDB[7];
   unchar cmnd8 = myqueue->req.bCDB[8];
/*when use_sg=0,tranfer address to prdt[]**************

   if(myqueue->req.bSG_Count==0){
	myqueue->prdt[0].wStartAddr=wV2P2(myqueue->req.p_buf);
	myqueue->prdt[0].wCount=(((cmnd7 << 8) | cmnd8 ) * 512) |(1<<31);
	myqueue->req.p_sg=&myqueue->prdt[0];	
   }
end*/
   switch( myqueue->req.bCDB[0] ) {
	case scsiREAD_6  :		//	0x08
	case scsiWRITE_6 :		//	0x0A
	   if(myqueue->req.bSG_Count==0){
		myqueue->prdt[0].wStartAddr=wV2P2(myqueue->req.p_buf);
		myqueue->prdt[0].wCount=((cmnd4?(cmnd4*512):0x20000)) |(1<<31);
		myqueue->req.p_sg=&myqueue->prdt[0];	
   	   }	
   	   myqueue->req.wSize = ( cmnd4 ? (cmnd4 * 512) : 0x20000 );
	   if(!myqueue->req.bCDB[4]) PTI_printk("fasttrak_rw_cmd: cmd->cmnd[4] = 0\n ");
	   break;						// 256*512
	case scsiREAD_10  :		//	0x28
	case scsiWRITE_10 :		//	0x2A
	   if(myqueue->req.bSG_Count==0){
		myqueue->prdt[0].wStartAddr=wV2P2(myqueue->req.p_buf);
		myqueue->prdt[0].wCount=(((cmnd7 << 8) | cmnd8 ) * 512) |(1<<31);
		myqueue->req.p_sg=&myqueue->prdt[0];	
   	   }
	   myqueue->req.wSize = ( (cmnd7 << 8) | cmnd8 ) * 512 ;
	   break;
   }
   buf_submit(&(myqueue->req));
   return(0);
}
/***********************************************************
*  Functions: (Frank Tiernan)
*    int reset_controller(void) 
*
*  Description:
*    Reset the Engine. Called when Scsi_Cmnd times out at the
*  OS level by fasttrak_reset().And Set MaintenanceLength = 0
*  if reset_controller is called to slow down rebuild rate so
*  scsi command can complete.
***********************************************************/
extern BoOL SetRebuildRate(U8);
extern BoOL SetExpConvRate(U8);
int reset_controller(void) {
   int index;
   
   // If rebuild is in progress, the command may never complete in time
   // unless rebuild stops... so modify the rebuild rate of the engine
   bResetController = TRUE;  
   //MaintenanceLength = 0;
   SetRebuildRate(0);
   SetExpConvRate(0);
   
   for (index = 0; index <= Can_Queue; index++) {  
    	// Flush out all the pending requests. It is assumed one 
    	// is the one that timed out.
    	if (Queue[index].pending) {
      	   if (!RequestTimedOut((REQ *)&Queue[index])) {
           // we might have to do something here to wait
           // for pending commands that have not timed out
      	   }
    	}
  }

  return TRUE;
}

#ifdef _PBM_
extern void Bufhooked_RAID_Return(REQ *req);
extern void RAID_Return(bpREQ pRaidReq);
/*
void RAID_Return(bpREQ pRaidReq)
{
	Bufhooked_RAID_Return(pRaidReq);
}
*/
void Engine_RAID_Return(bpREQ request)
#else
/***********************************************************
*  Functions:
*    exported by the OS Driver  
*
*  Description:
*   
***********************************************************/
void RAID_Return(bpREQ request)
#endif
{
   REQ_Head_t 	*myqueue = (REQ_Head_t *) request ;
   void		*cmd  = myqueue -> cmd ;
   void		(*fn)(void *) = myqueue -> fn ;
   int		result ; 
   if ( request->bAdapterStatus != 1 ) 		// normal : bAdapterStatus = 1 
   {
   }
   if ( request->bCDB[0] == scsiINQUIRY ) {
	unchar	bus, fun, class=0;
	ushort	PromiseController = 0 ;
	while( PTI_DEVID[class] ) 
	   if( !PTI_pcibios_find_device(0x105A,PTI_DEVID[class++],0,&bus,&fun) )
	   	PromiseController = 1 ;
	if( !PromiseController ) {
	   request->bAdapterStatus = 0 ; 
	}
   }
   else if ( request->bCDB[0] == scsiREAD_CAPACITY ) 
   	ReadCapacityCompleted = 1 ;
   if( request->bAdapterStatus == 1 ) 
	result = DID_OK << 16 ;
   else if( request->bAdapterStatus == 0x84 ) {		// HDDs power off
   	DEB(printk("*** bAdapterStatus(0x84) - HDD Power Off ****\n"););
   	result = ( aryDef[myqueue->req.bID].blRAID ) ? 
	   ( DID_OK << 16 ) : ( DID_ABORT << 16 ) ; // Mirror : blRAID = 1
   }
   else 
	result = DID_ABORT << 16 ;
   

   SCSIRequestFlag -- ;
   PoolF->next = myqueue ;
   PoolF = myqueue ;
   PoolF->next = NULL ;
   EndRequest(cmd,fn,result);
   return ;
}
// -----------------------------------------------
void SubmitTimer(bpEVENT p_ev)
{
#if 1	
   DebEngFun(printk("DebEngFun : SubmitTimer(%d).\n",p_ev->hETA););
   PTI_del_timer((struct PTI_timer_list* )(&submit_timer));
   init_timer(&submit_timer);
   submit_timer.expires = PTI_get_jiffies() + (p_ev->hETA * 55 * HZ / 1000) ;

   submit_timer.data = ( ulong ) p_ev ;
   submit_timer.function = submit_timer_handler ;
   PTI_add_timer((struct PTI_timer_list* )(&submit_timer));

   hTicksTillEvent = p_ev->hETA ; 
#endif   
   return;
}
// -----------------------------------------------
ulong wGet55msecTicksSinceMidnight(void)
{
   DebEngFun(printk("DebEngFun : wGet55msecTicksSinceMidnight.\n"););
   return(PTI_get_jiffies());
}
// -----------------------------------------------
HALF hGetDaysSince1980(void)
{
   struct timeval	tv ;
   ulong		days ,seconds ;

   DebEngFun(printk("DebEngFun : hGetDaysSince1980(void)\n"););
   PTI_do_gettimeofday(&tv);				// from 1970
   seconds = (ulong)tv.tv_sec - (10*365+2)*24*60*60 ; 	// from 1980
   days = seconds / (24*60*60) ;
//   PTI_printk("hGetDaysSince1980(void) : %ld\n",days);
   return(days);
}
// -----------------------------------------------
void DisableInt(void)
{
   if(INestcli==0) {
	save_flags(SaveFlags);
	cli();
   }
   INestcli++;
   return;
}
// -----------------------------------------------
void EnableInt(void)
{
   if( INestcli==1 ) {
	restore_flags(SaveFlags);
   }
   INestcli--;
	if(INestcli==0xFF) {
		INestcli=0;
	}
   return;
} 
int GetCHS(unchar dev,uint *pcyls,uint *pheads,uint *psectors) {

   struct PTI_driveid	*id=PTI_drive[dev].identify;
   uint		cyls    = id->cyls ;
   uint		heads   = id->heads ;
   uint		sectors = id->sectors ;
   ushort	CapH = id->max_48LBA[1] ;
   ushort	CapL = id->max_48LBA[0] ;
   DebEngFun(ft_printk("DebEngFun : wReservedSectorLBA()\n"););
   if ( DiskInfo[dev/8][dev/2][dev%2].bSupport48bitLBA==1 ) {
	if( cyls == 16383 ) {
	   cyls = ( ((CapH << 16)+CapL) / heads ) / sectors ;
	   // Don't transfer CHS again when cyls larger then 65535, 
	   // in order to match BIOS setting on Reserved Sector LBA
//	   if ( cyls >= 65535 ) {
//	   	cyls = 65535;
//	   	heads = 255;
//	   	sectors = ( ((CapH << 16)+CapL) / cyls ) / heads;
//	   }
	}
   }	
   else {	   
	if( cyls == 16383 )
	   cyls = ( id->lba_capacity / heads ) / sectors ;
   }
   *pcyls    = cyls ;
   *pheads   = heads ;
   *psectors = sectors ;
   return(0);
}
ulong wGetGeometry(ByTE bCh,ByTE bID,ByTE flag) {

   unchar	dev = bCh*2 + bID ;
   uint		cyls, heads, sectors ;

   DebEngFun(ft_printk("DebEngFun : wGetGeometry().\n"););
   GetCHS(dev,&cyls,&heads,&sectors) ;
   return( (((ushort)cyls)<<16)+(heads<<8)+sectors);
}
// -----------------------------------------------
ulong wReservedSectorLBA(ByTE bCh,ByTE bID,ByTE flag)
{
   unchar	dev = bCh*2 + bID ;
   uint		cyls, heads, sectors ;
   
   DebEngFun(ft_printk("DebEngFun : wReservedSectorLBA()\n"););
   GetCHS(dev,&cyls,&heads,&sectors) ;
   if(flag)
	return (( (cyls-1)*heads - 1 ) * sectors) ;
   else 		// default flag = 0
	return (( cyls * heads - 1 ) * sectors) ;
} 

// -----------------------------------------------
void wGetDriveName(ByTE bCh,ByTE bID,ByTE *pName)
{
// In fact : Engine does not call the fun.
   DebEngFun(printk("DebEngFun : wGetDriveName().\n"););
   return;
}
// -----------------------------------------------
ulong wLBAnext(ByTE bCh,ByTE bID)
{
// In fact : Engine does not call the fun.
   DebEngFun(printk("DebEngFun : wLBAnext()\n"););
   return(0);
}
// -------------------------------------------------------------------
BoOL bCheckDriveStatus(ByTE bCh,ByTE bID)
{
   struct PTI_driveid	*id ;
   unchar 		dev , result ;
   ushort		base2;
   //ulong		flags;
   static unchar	hide = 0 ;
   DebEngFun(printk("DebEngFun : bCheckDriveStatus(%d/%d)\n",bCh,bID););
   dev = bCh * 2 + bID ;
   base2 = PTI_drive[dev].pchannel->base+2 ;	
   if( bCh >= PTI_Max_Ch )	// channel exist or not
   	return(0);
   id = PTI_drive[dev].identify ;
   result = identify_device(dev,id) ;
   switch(result) {
	case 0	: break;
	case 3  : PTI_printk("FastTrak : (Error) Irq has been disabled !!\n");
	case 1  : 
	case 2  : return(0);		// Device not Exist 	
	default : PTI_printk("FastTrak : identify_device() fail !!\n");
   }
   PTI_drive[dev].present = 1 ;
   DEB(printk("FastTrak : FIND NEW DRIVE !! Device %d\n",dev););
   if( DisMirrorSync ) {
	if( !hide ){
   	   log_event(4121, 0, dev, 0,0,0,0); 	// New Disk plugged in
	}
	hide = dev+1 ;
   	return(0);
   }
   if( dev != hide-1 ){
   	log_event(4121, 0, dev, 0,0,0,0); 	// New Disk plugged in
   }
   hide = 0 ;
   return(1);
}
// -----------------------------------------------
void OutputToSerialPort(apBUF p_b )
{
// In fact : Engine does not call the fun.
   DebEngFun(printk("DebEngFun : OutputToSerialPort()\n"););
   return;
}   
// --------------------------------- ( for NEW ENGINE )
WORD _call wReadLBA_Register(ByTE bCh,ByTE bID)
{
   DEB(printk("wReadLBA_Register()\n"););
   return(0);
}
// -----------------------------------------------
void Led_Turnon(ByTE bCh, ByTE bID)
{
   DEB(printk("Led_Turnon() (WIN/NT) : rainmker.c\n"););
   return;
}
// -----------------------------------------------
ulong _call FastTrakNotification (WORD NotifyType,...)
{
   DEB(printk("FastTrakNotification() (WIN/NT) : rainmker.c\n"););
   return(0);
}
void PTI_memset(void *addr,int i,long size)
{
	memset(addr,i,size);
}
