/*********************************************************************
* channel_wrapper.c
* 08/23/00 Frank Tiernan
* 	This file wraps the linux kernel calls used by the lower 
* channel driver of the RAID engine. This allows the pre-compiled 
* channel driver to be compiled under various flavors of the linux kernel 
* 	The channel driver also uses the following structures:
* timer_list in /usr/src/linux/include/linux/timer.h
* Functions inside the channel driver that use these structures have
* been given wrappers so that in the worst case scenario that these
* structures change and no new engine binaries are released it
* may be possible to convert the new structures into the old structures. 
* 10/30/00 Linus Chen
*	Adding some wrapper functions for os driver.
*********************************************************************/
 
#include <asm-i386/io.h>
#include <asm-i386/ptrace.h>
#include <asm-i386/delay.h>
#include <linux/sched.h>
#include <linux/types.h>  
//#include <linux/malloc.h>
#include <linux/blk.h>
#include <linux/pci.h>

#include "wrapper.h"
#include "cam_con.h"
#include "cam_macr.h"

/* used same as list_head in linux kernel 
struct 	PTI_list_head {
	struct PTI_list_head *next, *prev;
};

 used same as timer_list in linux kernel 
struct 	PTI_timer_list {
	struct PTI_list_head list;
	unsigned long expires;
	unsigned long data;
	void (*function)(unsigned long);
};
*/

//extern void PTI_handler(int irq, void *dev_id);

unchar PTI_inb (ushort port) {
  	return inb(port);
}

unsigned PTI_inw (ushort port) {
	return inw(port);
}

void PTI_outb (unchar value, ushort port) {
  	outb(value, port);
}

void PTI_outw (ushort value, ushort port) {
	outw(value, port);
}

void PTI_outl (ulong value, ushort port) {
  	outl(value, port);
}

//int PTI_printk(const char *fmt, va_list *argp) {
int 	PTI_printk (const char *fmt, ...){
	return printk(fmt);
}


int PTI_sprintf(char * buf, const char * fmt, ...) {
	va_list args;
	int i;
	va_start(args, fmt);
	i=vsprintf(buf, fmt, args);
	va_end(args);
	return i;
}

void PTI_init_timer(struct PTI_timer_list * timer) {
/* If necessary convert PTI_timer_list into timer_list */
  	init_timer((struct timer_list *)timer);
}

void PTI_add_timer(struct PTI_timer_list * timer) {
/* If necessary convert PTI_timer_list into timer_list */
  	add_timer((struct timer_list *)timer);
}

void PTI_del_timer(struct PTI_timer_list * timer) {
/* If necessary convert PTI_timer_list into timer_list */
  	del_timer((struct timer_list *)timer);
}
  
void PTI_udelay(unsigned long usecs) {
  	udelay(usecs);
}

unsigned long volatile PTI_get_jiffies(void) {
/* jiffies is the number of clock ticks since the OS was booted */
/* it is exported by the linux kernel */
  	return jiffies;
}

extern U32 ISR(U32 IRQ);
int 		CYC_ISR = 0 ;
extern int 	CYC_TIMER ;
void PTI_handler_wrapper(int irq, void *dev_id, struct pt_regs *regs) {
   	/* channel driver is system independent so it does not
      	   need pt_regs structure */
	ulong	smpflag;
//	if( CYC_ISR )
//	   printk("PTI_handler_wrapper : CYC_ISR = 1 \n");	
//	if( CYC_TIMER )
//	   printk("PTI_handler_wrapper : CYC_TIMER = 1 \n");	
	CYC_ISR = 1 ;
   	spin_lock_irqsave(&io_request_lock,smpflag);
   	ISR(irq);
   	spin_unlock_irqrestore(&io_request_lock,smpflag);
	CYC_ISR = 0 ;
}

void PTI_do_gettimeofday(struct timeval *tv) {
	do_gettimeofday(tv);
}

void *PTI_kmalloc(unsigned int size,int priority) {
	return(kmalloc(size,priority));
}

void *PTI_memcpy( void * to, const void * from, size_t n){
	return memcpy(to,from,n);
}

void PTI_kfree(void *obj) {
	kfree(obj);
}

int PTI_pcibios_find_device(ushort ven, ushort id, ushort index, unchar *bus, unchar *fun) {
	return( pcibios_find_device(ven, id, index, bus, fun) );
}

int PTI_pcibios_read_config_word(unchar bus, unchar fun, unchar where, ushort *ptr) {
	return( pcibios_read_config_word(bus, fun, where, ptr) );
}

