/****************************************************************************
* wrapper.h
* 08/23/00 Frank Tiernan
* This file wraps the linux kernel calls used by the lower channel driver
* of the RAID engine. This allows the pre-compiled channel driver to be
* compiled under various flavors of the linux kernel 
*
* The channel driver relies on the following kernel structures:
* pt_regs in /usr/src/linux/include/asm/i386/ptrace.h
* timer_list in /usr/src/linux/include/linux/timer.h
* Their associated structures PTI_pt_regs and PTI_timer_list are based
* on the 2.4.0-test6 kernels. If their structure changes, it will require
* the recompilation of the channel driver with the new structure
* definitions.
**************************************************************************/

//#include <asm/io.h>
#include <linux/time.h>
#include <asm/system.h>

#ifndef HZ
#define HZ 100
#endif

#ifndef NULL
#define NULL 0
//defined in <linux/types.h>
typedef unsigned char  	unchar; 
typedef unsigned short 	ushort;
typedef unsigned int	uint;
typedef unsigned long  	ulong; 
#endif

struct 	PTI_list_head {
	struct PTI_list_head *next, *prev;
};

struct 	PTI_timer_list {
	struct PTI_list_head list;
	ulong expires;
	ulong data;
	void (*function)(ulong);
};

unchar 	PTI_inb (ushort port);
unsigned PTI_inw (ushort port);
void 	PTI_outb (unchar value, ushort port);
void 	PTI_outw (ushort value, ushort port);
void 	PTI_outl (ulong value, ushort port);
int 	PTI_printk (const char *fmt, ...);
int 	PTI_sprintf(char * buf, const char * fmt, ...); 


void 	PTI_init_timer(struct PTI_timer_list * timer);
void 	PTI_add_timer(struct PTI_timer_list * timer);
void 	PTI_del_timer(struct PTI_timer_list * timer);

void 	PTI_udelay(ulong usecs); 
void 	PTI_do_gettimeofday(struct timeval *tv);
ulong 	volatile PTI_get_jiffies(void);

//#define	PTI_cli()		cli()
//#define PTI_save_flags(x) 	save_flags(x)
//#define PTI_restore_flags(x) 	restore_flags(x)

void 	*PTI_kmalloc(uint size,int priority) ;
void 	PTI_kfree(void *obj) ;

int PTI_pcibios_find_device(ushort ven,ushort id,ushort index,unchar *bus,unchar *fun) ; 
int PTI_pcibios_read_config_word(unchar bus,unchar fun,unchar where,ushort *ptr) ;

